import warnings
from ibm_watsonx_data_integration.services.datastage.models.base_stage import BaseStage
from ibm_watsonx_data_integration.services.datastage.models.enums import PIVOT
from pydantic import Field
from typing import ClassVar


class pivot(BaseStage):

    op_name: ClassVar[str] = "PxPivot"
    node_type: ClassVar[str] = "execution_node"
    image: ClassVar[str] = "/data-intg/flows/graphics/palette/PxPivot.svg"
    label: ClassVar[str] = "Pivot Enterprise"
    runtime_column_propagation: bool = Field(False, alias="runtime_column_propagation")
    array_size: int | None = Field(1, alias="array_size")
    auto_column_propagation: bool | None = Field(None, alias="auto_column_propagation")
    buf_free_run: int | None = Field(50, alias="buf_free_run")
    buf_free_run_ronly: int | None = Field(50, alias="buf_free_run_ronly")
    buf_mode: PIVOT.BufMode | None = Field(PIVOT.BufMode.default, alias="buf_mode")
    buf_mode_ronly: PIVOT.BufModeRonly | None = Field(PIVOT.BufModeRonly.default, alias="buf_mode_ronly")
    coll_type: PIVOT.CollType | None = Field(PIVOT.CollType.auto, alias="coll_type")
    combinability: PIVOT.Combinability | None = Field(PIVOT.Combinability.auto, alias="combinability")
    current_output_link_type: str = Field("PRIMARY", alias="currentOutputLinkType")
    disk_write_inc: int | None = Field(1048576, alias="disk_write_inc")
    disk_write_inc_ronly: int | None = Field(1048576, alias="disk_write_inc_ronly")
    enable_flow_acp_control: bool = Field(True, alias="enableFlowAcpControl")
    enable_schemaless_design: bool = Field(False, alias="enableSchemalessDesign")
    execmode: PIVOT.Execmode | None = Field(PIVOT.Execmode.default_par, alias="execmode")
    flow_dirty: str | None = Field("false", alias="flow_dirty")
    hide: bool | None = Field(False, alias="hide")
    input_count: int | None = Field(0, alias="input_count")
    input_link_description: list | None = Field("", alias="inputLinkDescription")
    inputcol_properties: list | None = Field([], alias="inputcolProperties")
    key_col_select: PIVOT.KeyColSelect | None = Field(PIVOT.KeyColSelect.default, alias="keyColSelect")
    key_cols_coll: list | None = Field([], alias="keyColsColl")
    key_cols_coll_ordered: list | None = Field([], alias="keyColsCollOrdered")
    key_cols_coll_robin: list | None = Field([], alias="keyColsCollRobin")
    key_cols_none: list | None = Field([], alias="keyColsNone")
    key_cols_part: list | None = Field([], alias="keyColsPart")
    max_mem_buf_size: int | None = Field(3145728, alias="max_mem_buf_size")
    max_mem_buf_size_ronly: int | None = Field(3145728, alias="max_mem_buf_size_ronly")
    output_acp_should_hide: bool = Field(True, alias="outputAcpShouldHide")
    output_count: int | None = Field(0, alias="output_count")
    output_link_description: list | None = Field("", alias="outputLinkDescription")
    outputcol_properties: list | None = Field([], alias="outputcolProperties")
    part_client_dbname: str | None = Field(None, alias="part_client_dbname")
    part_client_instance: str | None = Field(None, alias="part_client_instance")
    part_dbconnection: str | None = Field("", alias="part_dbconnection")
    part_stable: bool | None = Field(None, alias="part_stable")
    part_stable_coll: bool | None = Field(False, alias="part_stable_coll")
    part_stable_ordered: bool | None = Field(False, alias="part_stable_ordered")
    part_stable_roundrobin_coll: bool | None = Field(False, alias="part_stable_roundrobin_coll")
    part_table: str | None = Field(None, alias="part_table")
    part_type: PIVOT.PartType | None = Field(PIVOT.PartType.auto, alias="part_type")
    part_unique: bool | None = Field(None, alias="part_unique")
    part_unique_coll: bool | None = Field(False, alias="part_unique_coll")
    part_unique_ordered: bool | None = Field(False, alias="part_unique_ordered")
    part_unique_roundrobin_coll: bool | None = Field(False, alias="part_unique_roundrobin_coll")
    perform_sort: bool | None = Field(False, alias="perform_sort")
    perform_sort_coll: bool | None = Field(False, alias="perform_sort_coll")
    perform_sort_modulus: bool | None = Field(False, alias="perform_sort_modulus")
    pivot_custom_input_o_s_h: str | None = Field("", alias="PivotCustomInputOSH")
    pivot_custom_o_s_h: str | None = Field("", alias="PivotCustomOSH")
    pivot_index: bool | None = Field(False, alias="pivotIndex")
    pivot_properties: list | None = Field([], alias="pivotProperties")
    pivot_type: PIVOT.PivotType | None = Field(PIVOT.PivotType.pivot, alias="pivot_type")
    preserve: PIVOT.Preserve | None = Field(PIVOT.Preserve.default_propagate, alias="preserve")
    queue_upper_size: int | None = Field(0, alias="queue_upper_size")
    queue_upper_size_ronly: int | None = Field(0, alias="queue_upper_size_ronly")
    runtime_column_propagation: bool | None = Field(None, alias="runtime_column_propagation")
    show_coll_type: int | None = Field(0, alias="showCollType")
    show_part_type: int | None = Field(1, alias="showPartType")
    show_sort_options: int | None = Field(0, alias="showSortOptions")
    sort_instructions: str | None = Field("", alias="sortInstructions")
    sort_instructions_text: str | None = Field("", alias="sortInstructionsText")
    stage_description: list | None = Field("", alias="stageDescription")
    vertical_pivot_custom_o_s_h: str | None = Field("", alias="VerticalPivotCustomOSH")

    def validate_parameters(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        (
            include.add("array_size")
            if (

                    self.pivot_type
                    and (
                        (hasattr(self.pivot_type, "value") and self.pivot_type.value == "verticalpivot")
                        or (self.pivot_type == "verticalpivot")
                    )

            )
            else exclude.add("array_size")
        )
        (
            include.add("vertical_pivot_custom_o_s_h")
            if (((not self.vertical_pivot_custom_o_s_h)) and (self.vertical_pivot_custom_o_s_h))
            else exclude.add("vertical_pivot_custom_o_s_h")
        )
        (
            include.add("pivot_custom_o_s_h")
            if (((not self.vertical_pivot_custom_o_s_h)) and (self.vertical_pivot_custom_o_s_h))
            else exclude.add("pivot_custom_o_s_h")
        )
        (
            include.add("pivot_custom_input_o_s_h")
            if (((not self.vertical_pivot_custom_o_s_h)) and (self.vertical_pivot_custom_o_s_h))
            else exclude.add("pivot_custom_input_o_s_h")
        )
        include.add("preserve") if (self.output_count and self.output_count > 0) else exclude.add("preserve")
        include.add("max_mem_buf_size") if (self.buf_mode != "nobuffer") else exclude.add("max_mem_buf_size")
        include.add("buf_free_run") if (self.buf_mode != "nobuffer") else exclude.add("buf_free_run")
        include.add("queue_upper_size") if (self.buf_mode != "nobuffer") else exclude.add("queue_upper_size")
        include.add("disk_write_inc") if (self.buf_mode != "nobuffer") else exclude.add("disk_write_inc")

        (
            include.add("runtime_column_propagation")
            if (not self.enable_schemaless_design)
            else exclude.add("runtime_column_propagation")
        )
        (
            include.add("auto_column_propagation")
            if (not self.output_acp_should_hide)
            else exclude.add("auto_column_propagation")
        )
        (
            include.add("max_mem_buf_size_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("max_mem_buf_size_ronly")
        )
        (
            include.add("buf_free_run_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("buf_free_run_ronly")
        )
        (
            include.add("queue_upper_size_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("queue_upper_size_ronly")
        )
        (
            include.add("disk_write_inc_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("disk_write_inc_ronly")
        )
        (
            include.add("part_stable")
            if (
                ((self.show_part_type))
                and (self.part_type != "auto")
                and (self.part_type != "db2connector")
                and (self.show_sort_options)
            )
            else exclude.add("part_stable")
        )
        (
            include.add("part_unique")
            if (
                ((self.show_part_type))
                and (self.part_type != "auto")
                and (self.part_type != "db2connector")
                and (self.show_sort_options)
            )
            else exclude.add("part_unique")
        )
        (
            include.add("key_cols_part")
            if (
                (
                    ((self.show_part_type))
                    and ((not self.show_coll_type))
                    and ((self.part_type != "auto"))
                    and ((self.part_type != "db2connector"))
                    and ((self.part_type != "modulus"))
                )
                or (
                    ((self.show_part_type))
                    and (not self.show_coll_type)
                    and (self.part_type == "modulus")
                    and (self.perform_sort_modulus)
                )
            )
            else exclude.add("key_cols_part")
        )
        (
            include.add("part_dbconnection")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_dbconnection")
        )
        (
            include.add("part_client_dbname")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_client_dbname")
        )
        (
            include.add("part_client_instance")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_client_instance")
        )
        (
            include.add("part_table")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_table")
        )
        (
            include.add("perform_sort")
            if (((self.show_part_type)) and (((self.part_type == "hash")) or (self.part_type == "range")))
            else exclude.add("perform_sort")
        )
        (
            include.add("perform_sort_modulus")
            if (((self.show_part_type)) and (self.part_type == "modulus"))
            else exclude.add("perform_sort_modulus")
        )
        (
            include.add("key_col_select")
            if (
                ((self.show_part_type))
                and (self.part_type == "modulus")
                and (not self.perform_sort_modulus)
            )
            else exclude.add("key_col_select")
        )
        (
            include.add("sort_instructions")
            if (
                ((self.show_part_type))
                and (
                    ((self.part_type == "db2part"))
                    or (self.part_type == "entire")
                    or (self.part_type == "random")
                    or (self.part_type == "roundrobin")
                    or (self.part_type == "same")
                )
            )
            else exclude.add("sort_instructions")
        )
        (
            include.add("sort_instructions_text")
            if (
                ((self.show_part_type))
                and (
                    ((self.part_type == "db2part"))
                    or (self.part_type == "entire")
                    or (self.part_type == "random")
                    or (self.part_type == "roundrobin")
                    or (self.part_type == "same")
                )
            )
            else exclude.add("sort_instructions_text")
        )
        include.add("coll_type") if (self.show_coll_type) else exclude.add("coll_type")
        include.add("part_type") if (self.show_part_type) else exclude.add("part_type")
        (
            include.add("perform_sort_coll")
            if (
                (
                    ((self.show_coll_type))
                    and (
                        ((self.coll_type == "ordered"))
                        or ((self.coll_type == "roundrobin_coll"))
                        or ((self.coll_type == "sortmerge"))
                    )
                )
                or (((not self.show_part_type)) and (not self.show_coll_type))
            )
            else exclude.add("perform_sort_coll")
        )
        (
            include.add("key_cols_coll")
            if (
                ((self.show_coll_type))
                and (not self.show_part_type)
                and (self.coll_type != "auto")
                and (((self.coll_type == "sortmerge")) or (self.perform_sort_coll))
            )
            else exclude.add("key_cols_coll")
        )
        (
            include.add("key_cols_none")
            if (
                ((not self.show_part_type))
                and (not self.show_coll_type)
                and (self.perform_sort_coll)
            )
            else exclude.add("key_cols_none")
        )
        (
            include.add("part_stable_coll")
            if (
                (((self.perform_sort_coll)))
                and (
                    (
                        ((not self.show_part_type))
                        and ((self.show_coll_type))
                        and ((self.coll_type != "auto"))
                        and ((self.show_sort_options))
                    )
                    or (
                        ((not self.show_part_type))
                        and (not self.show_coll_type)
                        and (self.show_sort_options)
                    )
                )
            )
            else exclude.add("part_stable_coll")
        )
        (
            include.add("part_unique_coll")
            if (
                (((self.perform_sort_coll)))
                and (
                    (
                        ((not self.show_part_type))
                        and ((self.show_coll_type))
                        and ((self.coll_type != "auto"))
                        and ((self.show_sort_options))
                    )
                    or (
                        ((not self.show_part_type))
                        and (not self.show_coll_type)
                        and (self.show_sort_options)
                    )
                )
            )
            else exclude.add("part_unique_coll")
        )
        return include, exclude

    def get_input_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = {"runtime_column_propagation"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_input_cardinality(self) -> dict:
        return {"min": 1, "max": 1}

    def get_output_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = set()
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_output_cardinality(self) -> dict:
        return {"min": 1, "max": 1}

    def get_parameters_props(self) -> dict:
        include, exclude = self.validate()
        props = {
            "array_size",
            "auto_column_propagation",
            "buf_free_run",
            "buf_free_run_ronly",
            "buf_mode",
            "buf_mode_ronly",
            "coll_type",
            "combinability",
            "current_output_link_type",
            "disk_write_inc",
            "disk_write_inc_ronly",
            "enable_flow_acp_control",
            "enable_schemaless_design",
            "execmode",
            "flow_dirty",
            "hide",
            "input_count",
            "input_link_description",
            "inputcol_properties",
            "key_col_select",
            "key_cols_coll",
            "key_cols_coll_ordered",
            "key_cols_coll_robin",
            "key_cols_none",
            "key_cols_part",
            "max_mem_buf_size",
            "max_mem_buf_size_ronly",
            "output_acp_should_hide",
            "output_count",
            "output_link_description",
            "outputcol_properties",
            "part_client_dbname",
            "part_client_instance",
            "part_dbconnection",
            "part_stable",
            "part_stable_coll",
            "part_stable_ordered",
            "part_stable_roundrobin_coll",
            "part_table",
            "part_type",
            "part_unique",
            "part_unique_coll",
            "part_unique_ordered",
            "part_unique_roundrobin_coll",
            "perform_sort",
            "perform_sort_coll",
            "perform_sort_modulus",
            "pivot_custom_input_o_s_h",
            "pivot_custom_o_s_h",
            "pivot_index",
            "pivot_properties",
            "pivot_type",
            "preserve",
            "queue_upper_size",
            "queue_upper_size_ronly",
            "runtime_column_propagation",
            "show_coll_type",
            "show_part_type",
            "show_sort_options",
            "sort_instructions",
            "sort_instructions_text",
            "stage_description",
            "vertical_pivot_custom_o_s_h",
        }
        required = {
            "current_output_link_type",
            "enable_flow_acp_control",
            "enable_schemaless_design",
            "output_acp_should_hide",
            "pivot_type",
        }
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_app_data_props(self) -> dict:
        return {
            "datastage": {
                "maxRejectOutputs": 0,
                "minRejectOutputs": 0,
                "maxReferenceInputs": 0,
                "minReferenceInputs": 0,
            }
        }



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import IBM_MQ_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class ClientChannelDefinitionTransportType(Enum):
    tcp = "tcp"
    udp = "udp"


class IbmMqConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "21364ca9-5b2d-323e-bd4d-59ba961f75fb"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    cas_lite_service_authorization_header: str | None = Field(None, alias="cas_lite_auth_header")
    channel_name: str | None = Field(None, alias="client_channel_definition.channel_name")
    connection_name: str | None = Field(None, alias="client_channel_definition.connection_name")
    transport_type: IBM_MQ_CONNECTION.ClientChannelDefinitionTransportType | None = Field(
        IBM_MQ_CONNECTION.ClientChannelDefinitionTransportType.tcp, alias="client_channel_definition.transport_type"
    )
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    password: str | None = Field(None, alias="password")
    queue_manager_name: str | None = Field(None, alias="queue_manager_name")
    cipher_spec: str | None = Field("ECDHE_RSA_AES_128_CBC_SHA256", alias="ssl_cipher_spec")
    client_ssl_certificate: str | None = Field(None, alias="ssl_client_certificate")
    client_ssl_key: str | None = Field(None, alias="ssl_client_key")
    ssl_connection: bool | None = Field(False, alias="ssl_connection")
    server_ssl_certificate: str | None = Field(None, alias="ssl_server_certificate")
    use_cas_lite_service: bool | None = Field(True, alias="use_cas_lite")
    username: str | None = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        include.add("password") if (not self.defer_credentials) else exclude.add("password")
        include.add("cipher_spec") if (self.ssl_connection) else exclude.add("cipher_spec")
        include.add("client_ssl_key") if (self.ssl_connection) else exclude.add("client_ssl_key")
        (
            include.add("use_cas_lite_service")
            if (self.cas_lite_service_authorization_header)
            else exclude.add("use_cas_lite_service")
        )
        (
            include.add("cas_lite_service_authorization_header")
            if (self.use_cas_lite_service)
            else exclude.add("cas_lite_service_authorization_header")
        )
        include.add("client_ssl_certificate") if (self.ssl_connection) else exclude.add("client_ssl_certificate")
        include.add("server_ssl_certificate") if (self.ssl_connection) else exclude.add("server_ssl_certificate")
        include.add("username") if (not self.defer_credentials) else exclude.add("username")
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")

        (
            include.add("use_cas_lite_service")
            if (self.cas_lite_service_authorization_header)
            else exclude.add("use_cas_lite_service")
        )
        (
            include.add("cas_lite_service_authorization_header")
            if (self.use_cas_lite_service == "true" or self.use_cas_lite_service)
            else exclude.add("cas_lite_service_authorization_header")
        )
        (
            include.add("server_ssl_certificate")
            if (self.ssl_connection == "true" or self.ssl_connection)
            else exclude.add("server_ssl_certificate")
        )
        (
            include.add("client_ssl_key")
            if (self.ssl_connection == "true" or self.ssl_connection)
            else exclude.add("client_ssl_key")
        )
        include.add("password") if (not self.defer_credentials) else exclude.add("password")
        include.add("username") if (not self.defer_credentials) else exclude.add("username")
        (
            include.add("client_ssl_certificate")
            if (self.ssl_connection == "true" or self.ssl_connection)
            else exclude.add("client_ssl_certificate")
        )
        (
            include.add("cipher_spec")
            if (self.ssl_connection == "true" or self.ssl_connection)
            else exclude.add("cipher_spec")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



import ibm_watsonx_data_integration.services.datastage.models.schema.field as SchemaField
import json
from ibm_watsonx_data_integration.services.datastage.models.enums import FIELD
from ibm_watsonx_data_integration.services.datastage.models.flow_json_model import FieldModelComplex
from ibm_watsonx_data_integration.services.datastage.models.schema import field
from ibm_watsonx_data_integration.services.datastage.models.schema.field.base_field import BaseField
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
)
from typing import Literal, Optional, overload

_ODBC_TYPE_TO_FIELD_CLASS = {
    "BIGINT": field.BigInt,
    "BINARY": field.Binary,
    "BIT": field.Bit,
    "CHAR": field.Char,
    "DATE": field.Date,
    "DECIMAL": field.Decimal,
    "DOUBLE": field.Double,
    "FLOAT": field.Float,
    "INTEGER": field.Integer,
    "LONGVARBINARY": field.LongVarBinary,
    "LONGVARCHAR": field.LongVarChar,
    "NUMERIC": field.Numeric,
    "REAL": field.Real,
    "SMALLINT": field.SmallInt,
    "TIME": field.Time,
    "TIMESTAMP": field.Timestamp,
    "TINYINT": field.TinyInt,
    "UNKNOWN": field.Unknown,
    "VARBINARY": field.VarBinary,
    "VARCHAR": field.VarChar,
    "NCHAR": field.NChar,
    "LONGNVARCHAR": field.LongNVarChar,
    "NVARCHAR": field.NVarChar,
}


class DataDefinitionMetadata(BaseModel):
    """System metadata about a table definition.

    :param str name: table definition name.
    :param str description: table definition description.
    """

    name: str = Field()
    description: str = Field("")

    @classmethod
    def from_dict(cls, _dict: dict) -> "DataDefinitionMetadata":
        """Initialize a DataDefinitionMetadata object from a json dictionary."""
        args = {}
        if (name := _dict.get("name")) is not None:
            args["name"] = name
        else:
            raise ValueError("Required property 'name' not present in DataDefinitionMetadata JSON")
        if (description := _dict.get("description")) is not None:
            args["description"] = description
        else:
            args["description"] = ""
        return cls(**args)


class DataDefinitionDataAsset(BaseModel):
    """column definitions and table properties.

    :param str mime_type: (optional)
    :param bool dataset: (optional)
    :param list[FieldModel] columns:
    :param object additional_properties: table properties.
    """

    columns: list[FieldModelComplex] = Field([])
    additional_properties: dict = Field({})
    mime_type: str | None = Field(None)
    dataset: bool | None = Field(None)

    @classmethod
    def from_dict(cls, _dict: dict) -> "DataDefinitionDataAsset":
        """Initialize a datadefinitiondataasset object from a json dictionary."""
        args = {}
        if (mime_type := _dict.get("mime_type")) is not None:
            args["mime_type"] = mime_type
        if (dataset := _dict.get("dataset")) is not None:
            args["dataset"] = dataset
        if (columns := _dict.get("columns")) is not None:
            args["columns"] = [FieldModelComplex.from_data_definition_column_dict(v) for v in columns]
        else:
            raise ValueError("Required property 'columns' not present in DataDefinitionDataAsset JSON")
        if (additional_properties := _dict.get("additionalProperties")) is not None:
            args["additional_properties"] = additional_properties
        else:
            raise ValueError("Required property 'additionalProperties' not present in DataDefinitionDataAsset JSON")
        return cls(**args)


class DataDefinitionTypeDefaults(BaseModel):
    """default properties for data types.

    :param dict general: (optional)
    :param dict string: (optional)
    :param dict decimal: (optional)
    :param dict numeric: (optional)
    :param dict date: (optional)
    :param dict time: (optional)
    :param dict timestamp: (optional)
    """

    general: dict | None = (None,)
    string: dict | None = (None,)
    decimal: dict | None = (None,)
    numeric: dict | None = (None,)
    date: dict | None = (None,)
    time: dict | None = (None,)
    timestamp: dict | None = (None,)

    @classmethod
    def from_dict(cls, _dict: dict) -> "DataDefinitionTypeDefaults":
        """Initialize a DataDefinitionTypeDefaults object from a json dictionary."""
        args = {}
        if (general := _dict.get("general")) is not None:
            args["general"] = general
        if (string := _dict.get("string")) is not None:
            args["string"] = string
        if (decimal := _dict.get("decimal")) is not None:
            args["decimal"] = decimal
        if (numeric := _dict.get("numeric")) is not None:
            args["numeric"] = numeric
        if (date := _dict.get("date")) is not None:
            args["date"] = date
        if (time := _dict.get("time")) is not None:
            args["time"] = time
        if (timestamp := _dict.get("timestamp")) is not None:
            args["timestamp"] = timestamp
        return cls(**args)


class DataDefinitionDSInfo(BaseModel):
    """data type defaults and format properties.

    :param List[object] data_types: (optional) definitions for custom data type.
    :param dict record_level: (optional)
    :param dict field_defaults: (optional)
    :param DataDefinitionTypeDefaults type_defaults: (optional) default properties
          for data types.
    """

    data_types: list[object] | None = (None,)
    record_level: dict | None = (None,)
    field_defaults: dict | None = (None,)
    type_defaults: Optional["DataDefinitionTypeDefaults"] = (None,)

    @classmethod
    def from_dict(cls, _dict: dict) -> "DataDefinitionDSInfo":
        """Initialize a DataDefinitionDSInfo object from a json dictionary."""
        args = {}
        if (data_types := _dict.get("data_types")) is not None:
            args["data_types"] = data_types
        if (record_level := _dict.get("record_level")) is not None:
            args["record_level"] = record_level
        if (field_defaults := _dict.get("field_defaults")) is not None:
            args["field_defaults"] = field_defaults
        if (type_defaults := _dict.get("type_defaults")) is not None:
            args["type_defaults"] = DataDefinitionTypeDefaults.from_dict(type_defaults)
        return cls(**args)


class DirectoryAsset(BaseModel):
    """DirectoryAsset.

    :param str path: (optional) The directory asset id.
    """

    path: str | None = None

    @classmethod
    def from_dict(cls, _dict: dict) -> "DirectoryAsset":
        """Initialize a DirectoryAsset object from a json dictionary."""
        args = {}
        if (path := _dict.get("path")) is not None:
            args["path"] = path
        return cls(**args)


class DataDefinitionEntity(BaseModel):
    """The underlying table definition.

    :param DataDefinitionDataAsset data_asset: column definitions and table
          properties.
    :param dict column_info:
    :param dict data_definition:
    :param DataDefinitionDSInfo ds_info: data type defaults and format properties.
    :param DirectoryAsset directory_asset: (optional)
    """

    data_asset: "DataDefinitionDataAsset"
    column_info: dict
    data_definition: dict
    ds_info: "DataDefinitionDSInfo"
    directory_asset: Optional["DirectoryAsset"] = None

    @classmethod
    def from_dict(cls, _dict: dict) -> "DataDefinitionEntity":
        """Initialize a DataDefinitionEntity object from a json dictionary."""
        args = {}
        if (data_asset := _dict.get("data_asset")) is not None:
            args["data_asset"] = DataDefinitionDataAsset.from_dict(data_asset)
        else:
            raise ValueError("Required property 'data_asset' not present in DataDefinitionEntity JSON")
        if (column_info := _dict.get("column_info")) is not None:
            args["column_info"] = column_info
        else:
            raise ValueError("Required property 'column_info' not present in DataDefinitionEntity JSON")
        if (data_definition := _dict.get("data_definition")) is not None:
            args["data_definition"] = data_definition
        else:
            raise ValueError("Required property 'data_definition' not present in DataDefinitionEntity JSON")
        if (ds_info := _dict.get("ds_info")) is not None:
            args["ds_info"] = DataDefinitionDSInfo.from_dict(ds_info)
        else:
            raise ValueError("Required property 'ds_info' not present in DataDefinitionEntity JSON")
        if (directory_asset := _dict.get("directory_asset")) is not None:
            args["directory_asset"] = DirectoryAsset.from_dict(directory_asset)
        return cls(**args)


ODBC_TO_CLASS = {
    "BIGINT": SchemaField.BigInt,
    "BINARY": SchemaField.Binary,
    "BIT": SchemaField.Bit,
    "CHAR": SchemaField.Char,
    "DATE": SchemaField.Date,
    "DECIMAL": SchemaField.Decimal,
    "DOUBLE": SchemaField.Double,
    "FLOAT": SchemaField.Float,
    "INTEGER": SchemaField.Integer,
    "LONGVARBINARY": SchemaField.LongVarBinary,
    "LONGVARCHAR": SchemaField.LongVarChar,
    "LONGNVARCHAR": SchemaField.LongVarChar,
    "NUMERIC": SchemaField.Numeric,
    "REAL": SchemaField.Real,
    "SMALLINT": SchemaField.SmallInt,
    "TIME": SchemaField.Time,
    "TIMESTAMP": SchemaField.Timestamp,
    "TINYINT": SchemaField.TinyInt,
    "UNKNOWN": SchemaField.Unknown,
    "VARBINARY": SchemaField.VarBinary,
    "VARCHAR": SchemaField.VarChar,
    "NCHAR": SchemaField.NChar,
    "NVARCHAR": SchemaField.VarChar,
    "WCHAR": SchemaField.NChar,
    "WLONGVARCHAR": SchemaField.LongNVarChar,
    "WVARCHAR": SchemaField.NVarChar,
}


class DataDefinition(BaseModel):
    model_config = ConfigDict(populate_by_name=True, use_enum_values=True, serialize_by_alias=True)
    asset_id: str | None = None
    proj_id: str | None = None
    name: str
    description: str = ""
    additional_properties: dict = Field({}, alias="additionalProperties")
    mime_type: str = "application/json"
    dataset: bool = True
    column_info: dict = {}
    data_definition: dict = {}
    data_types: list = []

    # Record level extended properties
    record_level_fill_char: FIELD.Fill | None = Field(None, alias="fill")
    record_level_final_delimiter: FIELD.FinalDelim | str | None = Field(None, alias="final_delim")
    record_level_final_delimiter_string: str | None = Field(None, alias="final_delim_string")
    record_level_intact: str | None = Field(None, alias="intact")
    record_level_check_intact: bool | None = Field(None, alias="check_intact")
    record_level_record_delimiter: FIELD.RecordDelim | None = Field(None, alias="record_delim")
    record_level_record_delimiter_string: str | None = Field(None, alias="record_delim_string")
    record_level_record_length: FIELD.RecordLength | None = Field(None, alias="record_length")
    record_level_record_prefix: FIELD.RecordPrefix | None = Field(None, alias="record_prefix")
    record_level_record_type: FIELD.RecordType | None = Field(None, alias="type")
    record_level_record_format: FIELD.Format | None = Field(None, alias="format")

    # Field defaults extended properties
    field_defaults_actual_field_length: int | None = Field(None, alias="actual_length")
    field_defaults_delimiter: FIELD.FieldDelim | str | None = Field(None, alias="delim")
    field_defaults_delimiter_string: str | None = Field(None, alias="delim_string")
    field_defaults_null_field_length: int | None = Field(None, alias="null_length")
    field_defaults_null_field_value: str | None = Field(None, alias="null_field")
    field_defaults_null_field_value_separator: FIELD.ValueSeparator | None = Field(None, alias="value_separator")
    field_defaults_prefix_bytes: FIELD.Prefix | None = Field(None, alias="prefix")
    field_defaults_print_field: bool | None = Field(False, alias="print_field")
    field_defaults_quote: FIELD.Quote | None = Field(None, alias="quote")
    field_defaults_vector_prefix: FIELD.VectorPrefix | None = Field(None, alias="vector_prefix")

    # Type defaults extended properties
    ## General
    general_byte_order: FIELD.ByteOrder | None = Field(None, alias="byte_order")
    general_char_set: FIELD.CharSet | None = Field(
        None, alias="char_set"
    )  # internal values may need to be uppercased during serialization
    general_data_format: FIELD.DataFormat | None = Field(None, alias="data_format")
    general_max_width: int | None = Field(None, alias="max_width")
    general_pad_char: FIELD.PadChar | int | None = Field(
        None, alias="pad_char"
    )  # TODO: This padchar has the wrong internal values, figure out how to resolve
    general_width: int | None = Field(None, alias="width")

    ## String
    string_export_ebcdic_as_ascii: bool | None = Field(None, alias="export_ebcdic_as_ascii")
    string_import_ascii_as_ebcdic: bool | None = Field(None, alias="import_ascii_as_ebcdic")

    ## Decimal
    decimal_allow_all_zeros: bool | None = Field(None, alias="allow_all_zeros")
    decimal_separator: FIELD.DecimalSeparator | None = Field(None, alias="decimal_separator")
    decimal_packed: FIELD.DecimalPacked | None = Field(None, alias="packed_value")
    decimal_sign_position: FIELD.SignPosition | None = Field(None, alias="sign_position")
    decimal_packed_signed: bool | None = Field(None, alias="packed_signed")  # TODO: is true by default
    decimal_precision: int | None = Field(None, alias="precision")  # TODO: str(int) when serialized
    decimal_rounding: FIELD.Round | None = Field(None, alias="round")
    decimal_scale: int | None = Field(None, alias="scale")  # TODO: str(int) when serialized

    ## Numeric
    numeric_c_format: str | None = Field(None, alias="c_format")
    numeric_in_format: str | None = Field(None, alias="in_format")
    numeric_out_format: str | None = Field(None, alias="out_format")

    ## Date
    date_days_since: int | None = Field(None, alias="days_since")  # TODO: str(int) when serialized
    date_format_string: str | None = Field(None, alias="format_string")
    date_is_julian: bool | None = Field(None, alias="julian")  # TODO: default as False

    ## Time
    time_format_string: str | None = Field(None, alias="format_string")
    time_midnight_seconds: bool | None = Field(None, alias="midnight_seconds")  # TODO: default as false

    ## Timestamp
    timestamp_format_string: str | None = Field(None, alias="format_string")

    directory_asset_path: str | None = None
    selected_fields: list[str] = []
    columns: list[FieldModelComplex] = []

    def select_fields(self, field_names: list[str]):
        self.selected_fields.extend(field_names)
        return self

    def select_field(self, field_name: str):
        self.selected_fields.append(field_name)
        return self

    def _get_fields(self) -> list[BaseField]:
        return [
            ODBC_TO_CLASS[column.odbc_type](column.name)._create_from_model(model=column)
            for column in self.columns
            if column.name in self.selected_fields
        ]

    def _get_metadata_props(self):
        props = {"name", "description"}
        return self.model_dump(include=props)

    def _get_entity_props(self):
        props = {"column_info", "data_definition"}
        dumped = self.model_dump(include=props)
        dumped["data_asset"] = self._get_data_asset_props()
        dumped["ds_info"] = self._get_ds_info_props()

        return dumped

    def _get_data_asset_props(self):
        props = {"mime_type", "dataset", "additional_properties"}

        dumped = self.model_dump(include=props, by_alias=True)
        dumped["columns"] = self._get_columns()

        return dumped

    def _get_columns(self):
        return [column.serialize_to_data_definition_column() for column in self.columns]

    def _get_ds_info_props(self):
        props = {"data_types"}

        dumped = self.model_dump(include=props)
        dumped["record_level"] = self._get_record_level_props()
        dumped["field_defaults"] = self._get_field_defaults_props()
        dumped["type_defaults"] = self._get_type_defaults_props()

        return dumped

    def _get_record_level_props(self):
        props = {
            "record_level_fill_char",
            "record_level_final_delimiter",
            "record_level_final_delimiter_string",
            "record_level_record_delimiter",
            "record_level_record_delimiter_string",
            "record_level_record_length",
            "record_level_record_prefix",
        }
        dumped: dict = json.loads(self.model_dump_json(include=props, exclude_none=True))
        dumped["intact"] = self._get_intact_props()
        if not dumped["intact"]:
            dumped.pop("intact")
        dumped["record_format"] = self._get_record_format()
        if not dumped["record_format"]:
            dumped.pop("record_format")
        return dumped

    def _get_intact_props(self):
        props = {"record_level_intact", "record_level_check_intact"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_record_format(self):
        props = {"record_level_record_type", "record_level_record_format"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_field_defaults_props(self):
        props = {
            "field_defaults_actual_field_length",
            "field_defaults_delimiter",
            "field_defaults_delimiter_string",
            "field_defaults_null_field_length",
            "field_defaults_prefix_bytes",
            "field_defaults_print_field",
            "field_defaults_quote",
            "field_defaults_vector_prefix",
        }
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        dumped["null_field"] = self._get_null_field_props()
        if not dumped["null_field"]:
            dumped.pop("null_field")
        return dumped

    def _get_null_field_props(self):
        props = {
            "field_defaults_null_field_value",
            "field_defaults_null_field_value_separator",
        }
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_type_defaults_props(self):
        dumped = {}
        dumped["general"] = self._get_general_props()
        dumped["string"] = self._get_string_props()
        dumped["decimal"] = self._get_decimal_props()
        dumped["numeric"] = self._get_numeric_props()
        dumped["date"] = self._get_date_props()
        dumped["time"] = self._get_time_props()
        dumped["timestamp"] = self._get_timestamp_props()
        return dumped

    def _get_general_props(self):
        props = {
            "general_byte_order",
            "general_char_set",
            "general_data_format",
            "general_max_width",
            "general_pad_char",
            "general_width",
        }
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_string_props(self):
        props = {"string_export_ebcdic_as_ascii", "string_import_ascii_as_ebcdic"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_decimal_props(self):
        props = {
            "decimal_allow_all_zeros",
            "decimal_separator",
            "decimal_packed",
            "decimal_sign_position",
            "decimal_packed_signed",
            "decimal_precision",
            "decimal_rounding",
            "decimal_scale",
        }
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_numeric_props(self):
        props = {"numeric_c_format", "numeric_in_format", "numeric_out_format"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_date_props(self):
        props = {"date_days_since", "date_format_string", "date_is_julian"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_time_props(self):
        props = {"time_format_string", "time_midnight_seconds"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def _get_timestamp_props(self):
        props = {"timestamp_format_string"}
        dumped = json.loads(self.model_dump_json(include=props, exclude_none=True))
        return dumped

    def from_dict(properties: dict) -> "DataDefinition":
        metadata = DataDefinitionMetadata.from_dict(properties["metadata"])
        entity = DataDefinitionEntity.from_dict(properties["entity"])
        if entity.directory_asset is None:
            director_asset_path = None
        else:
            director_asset_path = entity.directory_asset.path

        def wrap_in_enum(value, enum_class):
            return enum_class(value) if value is not None else None

        record_level = entity.ds_info.record_level
        if record_level.get("intact"):
            record_level_intact = record_level["intact"].get("intact")
            record_level_check_intact = record_level["intact"].get("check_intact")
        else:
            record_level_intact = None
            record_level_check_intact = None
        if record_level.get("final_delim") == "\\\\t":
            record_level_final_delimiter = r"\\t"
        else:
            record_level_final_delimiter = wrap_in_enum(record_level.get("final_delim"), FIELD.FinalDelim)

        if record_level.get("record_format"):
            record_level_record_format = wrap_in_enum(record_level["record_format"].get("format"), FIELD.Format)
            record_level_record_type = wrap_in_enum(record_level["record_format"].get("type"), FIELD.RecordType)
        else:
            record_level_record_format = None
            record_level_record_type = None

        field_defaults = entity.ds_info.field_defaults

        if field_defaults.get("null_field") and isinstance(field_defaults.get("null_field"), dict):
            field_defaults_null_field_value = field_defaults["null_field"].get("null_field")
            field_defaults_null_field_value_separator = wrap_in_enum(
                field_defaults["null_field"].get("value_separator"),
                FIELD.ValueSeparator,
            )
        elif field_defaults.get("null_field") and isinstance(field_defaults.get("null_field"), str):
            field_defaults_null_field_value = field_defaults.get("null_field")
            field_defaults_null_field_value_separator = None
        else:
            field_defaults_null_field_value = None
            field_defaults_null_field_value_separator = None

        if field_defaults.get("delim") == "\\\\t":
            field_defaults_delimiter = r"\\t"
        else:
            field_defaults_delimiter = wrap_in_enum(field_defaults.get("delim"), FIELD.Delim)

        general = entity.ds_info.type_defaults.general
        string = entity.ds_info.type_defaults.string
        decimal = entity.ds_info.type_defaults.decimal
        numeric = entity.ds_info.type_defaults.numeric
        date = entity.ds_info.type_defaults.date
        time_dict = entity.ds_info.type_defaults.time
        timestamp = entity.ds_info.type_defaults.timestamp

        return DataDefinition(
            name=metadata.name,
            description=metadata.description,
            additional_properties=entity.data_asset.additional_properties,
            mime_type=entity.data_asset.mime_type,
            dataset=entity.data_asset.dataset,
            column_info=entity.column_info,
            data_definition=entity.data_definition,
            data_types=entity.ds_info.data_types,
            record_level_fill_char=wrap_in_enum(record_level.get("fill"), FIELD.Fill),
            record_level_final_delimiter=record_level_final_delimiter,
            record_level_final_delimiter_string=record_level.get("final_delim_string"),
            record_level_intact=record_level_intact,
            record_level_check_intact=record_level_check_intact,
            record_level_record_delimiter=wrap_in_enum(
                (record_level.get("record_delim").replace("\\n", "\n") if record_level.get("record_delim") else None),
                FIELD.RecordDelim,
            ),
            record_level_record_delimiter_string=record_level.get("record_delim_string"),
            record_level_record_length=wrap_in_enum(record_level.get("record_length"), FIELD.RecordLength),
            record_level_record_prefix=wrap_in_enum(record_level.get("record_prefix"), FIELD.RecordPrefix),
            record_level_record_type=record_level_record_type,
            record_level_record_format=record_level_record_format,
            field_defaults_actual_field_length=field_defaults.get("actual_length"),
            field_defaults_delimiter=field_defaults_delimiter,
            field_defaults_delimiter_string=field_defaults.get("delim_string"),
            field_defaults_null_field_length=field_defaults.get("null_length"),
            field_defaults_null_field_value=field_defaults_null_field_value,
            field_defaults_null_field_value_separator=field_defaults_null_field_value_separator,
            field_defaults_prefix_bytes=wrap_in_enum(field_defaults.get("prefix"), FIELD.Prefix),
            field_defaults_print_field=field_defaults.get("print_field"),
            field_defaults_quote=wrap_in_enum(field_defaults.get("quote"), FIELD.Quote),
            field_defaults_vector_prefix=wrap_in_enum(field_defaults.get("vector_prefix"), FIELD.VectorPrefix),
            general_byte_order=wrap_in_enum(general.get("byte_order"), FIELD.ByteOrder),
            general_char_set=wrap_in_enum(general.get("char_set"), FIELD.CharSet),
            general_data_format=wrap_in_enum(general.get("data_format"), FIELD.DataFormat),
            general_max_width=general.get("max_width"),
            general_pad_char=wrap_in_enum(general.get("pad_char"), FIELD.PadChar),
            general_width=general.get("width"),
            string_export_ebcdic_as_ascii=string.get("export_ebcdic_as_ascii"),
            string_import_ascii_as_ebcdic=string.get("import_ascii_as_ebcdic"),
            decimal_allow_all_zeros=decimal.get("allow_all_zeros"),
            decimal_separator=wrap_in_enum(decimal.get("decimal_separator"), FIELD.DecimalSeparator),
            decimal_packed=wrap_in_enum(decimal.get("packed_value"), FIELD.DecimalPacked),
            decimal_sign_position=wrap_in_enum(decimal.get("sign_position"), FIELD.SignPosition),
            decimal_packed_signed=decimal.get("packed_signed"),
            decimal_precision=decimal.get("precision"),
            decimal_rounding=wrap_in_enum(decimal.get("round"), FIELD.Round),
            decimal_scale=decimal.get("scale"),
            numeric_c_format=numeric.get("c_format"),
            numeric_in_format=numeric.get("in_format"),
            numeric_out_format=numeric.get("out_format"),
            date_days_since=date.get("days_since"),
            date_format_string=date.get("format_string"),
            date_is_julian=date.get("julian"),
            time_format_string=time_dict.get("format_string"),
            time_midnight_seconds=time_dict.get("midnight_seconds"),
            timestamp_format_string=timestamp.get("format_string"),
            directory_asset_path=director_asset_path,
            columns=entity.data_asset.columns,
        )

    def set_record_level_check_intact(self, record_level_check_intact: bool):
        self.record_level_check_intact = record_level_check_intact
        return self

    def set_description(self, description: str):
        self.description = description
        return self

    def set_record_level_record_length(self, record_level_record_length: FIELD.RecordLength):
        self.record_level_record_length = record_level_record_length
        return self

    def set_numeric_c_format(self, numeric_c_format: str):
        self.numeric_c_format = numeric_c_format
        return self

    def set_date_is_julian(self, date_is_julian: bool):
        self.date_is_julian = date_is_julian
        return self

    def set_field_defaults_vector_prefix(self, field_defaults_vector_prefix: FIELD.VectorPrefix):
        self.field_defaults_vector_prefix = field_defaults_vector_prefix
        return self

    def set_decimal_sign_position(self, decimal_sign_position: FIELD.SignPosition):
        self.decimal_sign_position = decimal_sign_position
        return self

    def set_decimal_packed(self, decimal_packed: FIELD.DecimalPacked):
        self.decimal_packed = decimal_packed
        return self

    def set_field_defaults_print_field(self, field_defaults_print_field: bool):
        self.field_defaults_print_field = field_defaults_print_field
        return self

    def set_date_days_since(self, date_days_since: int):
        self.date_days_since = date_days_since
        return self

    def set_general_byte_order(self, general_byte_order: FIELD.ByteOrder):
        self.general_byte_order = general_byte_order
        return self

    def set_record_level_record_type(self, record_level_record_type: FIELD.RecordType):
        self.record_level_record_type = record_level_record_type
        return self

    def set_record_level_record_format(self, record_level_record_format: FIELD.Format):
        self.record_level_record_format = record_level_record_format
        return self

    def set_decimal_rounding(self, decimal_rounding: FIELD.Round):
        self.decimal_rounding = decimal_rounding
        return self

    def set_general_data_format(self, general_data_format: FIELD.DataFormat):
        self.general_data_format = general_data_format
        return self

    def set_record_level_final_delimiter_string(self, record_level_final_delimiter_string: str):
        self.record_level_final_delimiter_string = record_level_final_delimiter_string

        return self

    def set_timestamp_format_string(self, timestamp_format_string: str):
        self.timestamp_format_string = timestamp_format_string
        return self

    def set_string_import_ascii_as_ebcdic(self, string_import_ascii_as_ebcdic: bool):
        self.string_import_ascii_as_ebcdic = string_import_ascii_as_ebcdic
        return self

    def set_general_max_width(self, general_max_width: int):
        self.general_max_width = general_max_width
        return self

    def set_general_pad_char(self, general_pad_char: FIELD.PadChar):
        if general_pad_char.name == "space":
            general_pad_char = (
                32  # Handling for incompatiblility between existing PadChar enum value and PadChar value used here
            )
        self.general_pad_char = general_pad_char
        return self

    def set_field_defaults_delimiter(self, field_defaults_delimiter: FIELD.FieldDelim):
        self.field_defaults_delimiter = field_defaults_delimiter
        return self

    def set_field_defaults_prefix_bytes(self, field_defaults_prefix_bytes: FIELD.Prefix):
        self.field_defaults_prefix_bytes = field_defaults_prefix_bytes
        return self

    def set_decimal_scale(self, decimal_scale: int):
        self.decimal_scale = decimal_scale
        return self

    def set_time_midnight_seconds(self, time_midnight_seconds: bool):
        self.time_midnight_seconds = time_midnight_seconds
        return self

    def set_record_level_final_delimiter(self, record_level_final_delimiter: FIELD.FinalDelim):
        self.record_level_final_delimiter = record_level_final_delimiter
        return self

    def set_general_width(self, general_width: int):
        self.general_width = general_width
        return self

    def set_decimal_precision(self, decimal_precision: int):
        self.decimal_precision = decimal_precision
        return self

    def set_field_defaults_null_field_value_separator(
        self,
        field_defaults_null_field_value_separator: FIELD.ValueSeparator,
    ):
        self.field_defaults_null_field_value_separator = field_defaults_null_field_value_separator

        return self

    def set_general_char_set(self, general_char_set: FIELD.CharSet):
        self.general_char_set = general_char_set
        return self

    def set_record_level_record_delimiter_string(self, record_level_record_delimiter_string: str):
        self.record_level_record_delimiter_string = record_level_record_delimiter_string

        return self

    def set_field_defaults_actual_field_length(self, field_defaults_actual_field_length: int):
        self.field_defaults_actual_field_length = field_defaults_actual_field_length

        return self

    def set_numeric_in_format(self, numeric_in_format: str):
        self.numeric_in_format = numeric_in_format
        return self

    def set_numeric_out_format(self, numeric_out_format: str):
        self.numeric_out_format = numeric_out_format
        return self

    def set_string_export_ebcdic_as_ascii(self, string_export_ebcdic_as_ascii: bool):
        self.string_export_ebcdic_as_ascii = string_export_ebcdic_as_ascii
        return self

    def set_field_defaults_null_field_value(self, field_defaults_null_field_value: str):
        self.field_defaults_null_field_value = field_defaults_null_field_value
        return self

    def set_decimal_separator(self, decimal_separator: FIELD.DecimalSeparator):
        self.decimal_separator = decimal_separator
        return self

    def set_record_level_fill_char(self, record_level_fill_char: FIELD.Fill):
        self.record_level_fill_char = record_level_fill_char
        return self

    def set_decimal_allow_all_zeros(self, decimal_allow_all_zeros: bool):
        self.decimal_allow_all_zeros = decimal_allow_all_zeros
        return self

    def set_date_format_string(self, date_format_string: str):
        self.date_format_string = date_format_string
        return self

    def set_decimal_packed_signed(self, decimal_packed_signed: bool):
        self.decimal_packed_signed = decimal_packed_signed
        return self

    def set_field_defaults_delimiter_string(self, field_defaults_delimiter_string: str):
        self.field_defaults_delimiter_string = field_defaults_delimiter_string
        return self

    def set_field_defaults_quote(self, field_defaults_quote: FIELD.Quote):
        self.field_defaults_quote = field_defaults_quote
        return self

    def set_record_level_intact(self, record_level_intact: str):
        self.record_level_intact = record_level_intact
        return self

    def set_record_level_record_delimiter(self, record_level_record_delimiter: FIELD.RecordDelim):
        self.record_level_record_delimiter = record_level_record_delimiter
        return self

    def set_field_defaults_null_field_length(self, field_defaults_null_field_length: int):
        self.field_defaults_null_field_length = field_defaults_null_field_length

        return self

    def set_record_level_record_prefix(self, record_level_record_prefix: FIELD.RecordPrefix):
        self.record_level_record_prefix = record_level_record_prefix
        return self

    def set_time_format_string(self, time_format_string: str):
        self.time_format_string = time_format_string
        return self

    @overload
    def add_field(self, odbc_type: Literal["BIGINT"], name: str) -> field.BigInt: ...

    @overload
    def add_field(self, odbc_type: Literal["BINARY"], name: str) -> field.Binary: ...

    @overload
    def add_field(self, odbc_type: Literal["BIT"], name: str) -> field.Bit: ...

    @overload
    def add_field(self, odbc_type: Literal["CHAR"], name: str) -> field.Char: ...

    @overload
    def add_field(self, odbc_type: Literal["DATE"], name: str) -> field.Date: ...

    @overload
    def add_field(self, odbc_type: Literal["DECIMAL"], name: str) -> field.Decimal: ...

    @overload
    def add_field(self, odbc_type: Literal["DOUBLE"], name: str) -> field.Double: ...

    @overload
    def add_field(self, odbc_type: Literal["FLOAT"], name: str) -> field.Float: ...

    @overload
    def add_field(self, odbc_type: Literal["INTEGER"], name: str) -> field.Integer: ...

    @overload
    def add_field(self, odbc_type: Literal["LONGVARBINARY"], name: str) -> field.LongVarBinary: ...

    @overload
    def add_field(self, odbc_type: Literal["LONGVARCHAR"], name: str) -> field.LongVarChar: ...

    @overload
    def add_field(self, odbc_type: Literal["NUMERIC"], name: str) -> field.Numeric: ...

    @overload
    def add_field(self, odbc_type: Literal["REAL"], name: str) -> field.Real: ...

    @overload
    def add_field(self, odbc_type: Literal["SMALLINT"], name: str) -> field.SmallInt: ...

    @overload
    def add_field(self, odbc_type: Literal["TIME"], name: str) -> field.Time: ...

    @overload
    def add_field(self, odbc_type: Literal["TIMESTAMP"], name: str) -> field.Timestamp: ...

    @overload
    def add_field(self, odbc_type: Literal["TINYINT"], name: str) -> field.TinyInt: ...

    @overload
    def add_field(self, odbc_type: Literal["UNKNOWN"], name: str) -> field.Unknown: ...

    @overload
    def add_field(self, odbc_type: Literal["VARBINARY"], name: str) -> field.VarBinary: ...

    @overload
    def add_field(self, odbc_type: Literal["VARCHAR"], name: str) -> field.VarChar: ...

    @overload
    def add_field(self, odbc_type: Literal["NCHAR"], name: str) -> field.NChar: ...

    @overload
    def add_field(self, odbc_type: Literal["LONGNVARCHAR"], name: str) -> field.LongNVarChar: ...

    @overload
    def add_field(self, odbc_type: Literal["NVARCHAR"], name: str) -> field.NVarChar: ...

    def add_field(self, odbc_type: str, name: str):
        if odbc_type.upper() in _ODBC_TYPE_TO_FIELD_CLASS:
            field_class = _ODBC_TYPE_TO_FIELD_CLASS[odbc_type.upper()]
            new_field = field_class(name=name)
            self.columns.append(new_field.configuration)
            self.selected_fields.append(new_field.configuration.name)
            return new_field
        else:
            raise ValueError(
                f"Unsupported field type: {odbc_type}. Supported types are: {list(_ODBC_TYPE_TO_FIELD_CLASS.keys())}"
            )



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing the Resource Controller API client."""

import requests
from ibm_watsonx_data_integration.cpd_api.base import BaseAPIClient
from typing import TYPE_CHECKING
from urllib.parse import quote

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import BaseAuthenticator

DEFAULT_RESOURCE_CONTROLLER_API_VERSION = 2


class ResourceControllerApiClient(BaseAPIClient):
    """The API client of the Resource Controller service."""

    def __init__(
        self,
        auth: "BaseAuthenticator",  # pragma: allowlist secret
        base_url: str = "https://resource-controller.cloud.ibm.com",
    ) -> None:
        """The __init__ of the ResourceControllerApiClient.

        Args:
            auth: The Authentication object.
            base_url: The Cloud Pak for Data URL.
        """
        super().__init__(auth=auth, base_url=base_url)
        self.url_path_resource_instances = f"v{DEFAULT_RESOURCE_CONTROLLER_API_VERSION}/resource_instances"
        self.url_path_resource_groups = f"v{DEFAULT_RESOURCE_CONTROLLER_API_VERSION}/resource_groups"
        self.url_path_resource_keys = f"v{DEFAULT_RESOURCE_CONTROLLER_API_VERSION}/resource_keys"

    def get_resource_instances(self, params: dict) -> requests.Response:
        """Retrieves Resource Instances.

        Args:
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_resource_instances}"
        response = self.get(url=url, params=params)
        return response

    def create_resource_instance(self, data: dict) -> requests.Response:
        """Creates a Resource Instance.

        Args:
            data: The resource instance JSON.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_resource_instances}"
        response = self.post(url=url, data=data)
        return response

    def delete_resource_instance(self, resource_instance_id: str, recursive: bool) -> requests.Response:
        """Deletes a Resource Instance.

        Args:
            resource_instance_id: The resource instance ID.
            recursive: Whether to recursively delete keys of the resource instance.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_resource_instances}/{quote(resource_instance_id, safe='')}"
        params = {"recursive": recursive}
        response = self.delete(url=url, params=params)
        return response

    def get_resource_groups(self) -> requests.Response:
        """Gets the Resource Groups.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_resource_groups}"
        response = self.get(url=url)
        return response

    def delete_resource_keys(self, resource_instance_id: str) -> requests.Response:
        """Deletes the Resource Keys.

        Args:
            resource_instance_id: The resource instance ID.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_resource_keys}/{quote(resource_instance_id, safe='')}"
        response = self.delete(url=url)
        return response

    def get_resource_keys(self, resource_id: str) -> requests.Response:
        """Gets the Resource Keys.

        Args:
            resource_id: The resource ID.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_resource_keys}"
        params = {"resource_id": resource_id}
        response = self.get(url=url, params=params)
        return response



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""This module contains project-wide constants."""

from enum import Enum

DATASTAGE = "datastage"

RESOURCE_PLAN_ID_MAP = {DATASTAGE: "aaa6e83e-27df-4393-9a55-63feec09c685"}
DEFAULT_RESOURCE_REGION_ID_MAP = {DATASTAGE: "us-south"}

# BASE MODEL VARS
HIDDEN_DICTIONARY = "_hidden_key_paths"
EXPOSE_SUB_CLASS = "_expose"

# PROD URLS
PROD_BASE_API_URL = "https://api.dataplatform.cloud.ibm.com"
PROD_BASE_URL = "https://cloud.ibm.com"


class DataActions(str, Enum):
    """Actions supported in EXPOSED_DATH_PATH."""

    IGNORE = "IGNORE"
    EXPOSE = "EXPOSE"  # Shouldn't be used on the base level


SUPPORTED_FLOWS = {"streamsets", "datastage"}



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""This module defines a utility function to create JSON Patch format payloads.

Implementation is limited to the following use cases:
    1. Changing property values.
    2. Removing entries from a collection list.
    3. Adding entries to a collection list.
"""

import json
import jsondiff
from abc import ABC, abstractmethod
from collections.abc import MutableMapping


def prepare_json_patch_payload(origin: dict, updated: dict, exclude_properties: list[str] | None = None) -> str:
    """Entrypoint function for creating a JSON Patch format-compliant payload.

    Args:
        origin: Dictionary with original data before changes.
        updated: Dictionary after applying changes and updates.
        exclude_properties: List of flatten keys for properties to not include in payload.

    Returns:
        A JSON string compliant with the JSON Patch format standard.
    """
    difference = jsondiff.diff(origin, updated, syntax="explicit")
    flatten = flatten_jsondiff_explicit_result(difference)
    return create_json_patch_document(flatten, exclude_properties=exclude_properties)


def flatten_jsondiff_explicit_result(
    dictionary: MutableMapping, parent_key: str = "", separator: str = "/", last_operation: str = ""
) -> dict:
    """Flattens key paths from the result of ``jsondiff.diff``.

    This function is tightly coupled with the 3rd party library ``jsondiff``.
    It will work correctly for results returned by the ``explicit`` flag for the ``syntax`` argument.

    Args:
        dictionary: A result from ``jsondiff.diff`` execution.
        parent_key: Parent key from the perspective of the current processing key.
        separator: Character that will be used to delimit keys.
        last_operation: Operations mean keys starting with the ``$`` character.
            This parameter holds the last one in the current traversable path.

    Returns:
        A dictionary with flattened paths as keys and the original values as values.
    """
    items = []

    for key, value in dictionary.items():
        if str(key).startswith("$"):
            last_operation = key
            new_key = f"{parent_key}" if parent_key else ""
        else:
            new_key = f"{parent_key}{separator}{key}" if parent_key else key

        if isinstance(value, MutableMapping):
            items.extend(
                flatten_jsondiff_explicit_result(
                    value, new_key, separator=separator, last_operation=last_operation
                ).items()
            )
        else:
            items.append((f"{last_operation}{separator}{new_key}", value))
    return dict(items)


def create_json_patch_document(flatten_dict: dict, exclude_properties: list[str] | None = None) -> str:
    """Creates JSON document compliant with JSON Path format standard.

    Args:
        flatten_dict: Dictionary with flatted keys to values returned by ``flatten_jsondiff_explicit_result``.
        exclude_properties: List of flatten key paths for properties to exclude from patch document.

    Returns:
        A string representing JSON Path format document.
    """
    if exclude_properties is None:
        exclude_properties = list()

    exclude_properties_lower = list(map(lambda s: s.lower(), exclude_properties))
    operations = []

    for k, v in flatten_dict.items():
        path = k[k.find("/") :]

        if path.lower() in exclude_properties_lower:
            continue

        operation = map_operations(k[: k.find("/")], path, v)
        operations.extend(operation.value)

    return json.dumps(sorted(operations, key=operation_weight_sort_key))


class Operation(ABC):
    """Represent single operation acceptable by JSON Patch Format."""

    def __init__(self, path: str, raw_data: list | dict | object) -> None:
        """Initialize Operation class with private attributes.

        Args:
            path: Flatten key path.
            raw_data: Value for given key, format defined by ``json.diff``.
        """
        self._path = path
        self._raw_data = raw_data

    @property
    @abstractmethod
    def value(self) -> list[dict]:
        """Returns list with JSON Path format operation objects."""
        raise NotImplementedError


class AddOperation(Operation):
    """Creates operations for adding objects."""

    @property
    def value(self) -> list[dict]:
        """Returns list with JSON Path format operation objects."""
        result = []
        if isinstance(self._raw_data, list):
            for add_object in self._raw_data:
                index = add_object[0]
                value = add_object[1]
                result.append({"op": "add", "path": f"{self._path}/{index}", "value": value})
        else:
            result.append(
                {
                    "op": "add",
                    "path": self._path,
                    "value": self._raw_data,
                }
            )
        return result


class RemoveOperation(Operation):
    """Creates operations for removing objects."""

    @property
    def value(self) -> list[dict]:
        """Returns list with JSON Path format operation objects."""
        result = []
        if isinstance(self._raw_data, list):
            for index in self._raw_data:
                result.append({"op": "remove", "path": f"{self._path}/{index}"})
        else:
            result.append(
                {
                    "op": "remove",
                    "path": self._path,
                }
            )
        return result


class ReplaceOperation(Operation):
    """Creates operations for replacing objects."""

    @property
    def value(self) -> list[dict]:
        """Returns list with JSON Path format operation objects."""
        result = [{"op": "replace", "path": self._path, "value": self._raw_data}]
        return result


def map_operations(op_name: str, path: str, raw_data: dict | list | object) -> Operation:
    """Maps an operation from the ``jsondiff`` library to  corresponding in the JSON Path format specification.

    Args:
        op_name: Operation name used in ``jsondiff`` library, prefixed with ``$``.
        path: Flatten key path.
        raw_data: Operation related information created by ``jsondiff`` library.

    Returns:
        An Operation object that performs the creation of the appropriate operation object.

    Raises:
        ValueError: If an undefined operation is provided.
    """
    op_name_lower = op_name.lower()

    if op_name_lower == "$update":
        return ReplaceOperation(path, raw_data)
    elif op_name_lower == "$insert":
        return AddOperation(path, raw_data)
    elif op_name_lower == "$delete":
        return RemoveOperation(path, raw_data)
    else:
        raise ValueError("Undefined operation!")


def operation_weight_sort_key(operation: dict) -> int:
    """Controls the order of operation execution.

    Args:
        operation: A JSON Path format operation dictionary.

    Returns:
        A weight for the given operation, which will be used when sorting.
    """
    op_weight_map = {
        "remove": 1,
        "add": 2,
        "replace": 3,
    }
    return op_weight_map[operation["op"]]



# -------------------------------------------------------------------
# This file was auto-generated. Do not edit directly.
# Generated by: StageGeneratorV2
# -------------------------------------------------------------------

from typing import TYPE_CHECKING

from ibm_watsonx_data_integration.services.datastage.models.flow.dag import StageNode
from ibm_watsonx_data_integration.services.datastage.models.stage_models.apache_kafka import apache_kafka

from ibm_watsonx_data_integration.cpd_models.connections_model import Connection
from ibm_watsonx_data_integration.services.datastage.models.connections.apache_kafka_connection import ApacheKafkaConn


if TYPE_CHECKING:
    from ibm_watsonx_data_integration.services.datastage.models.flow.dag import DAG



class ApacheKafkaStage(StageNode):
    def __init__(self, dag: "DAG", configuration: apache_kafka=None, *, label: str = None):
        super().__init__(dag, configuration=configuration, label=label)
        self.configuration: apache_kafka = configuration or apache_kafka()

    def use_connection(self, connection: Connection):
        self.configuration.connection = ApacheKafkaConn().from_connection(connection)



import warnings
from ibm_watsonx_data_integration.services.datastage.models.base_stage import BaseStage
from ibm_watsonx_data_integration.services.datastage.models.connections.sapbulkextract_connection import (
    SapbulkextractConn,
)
from ibm_watsonx_data_integration.services.datastage.models.enums import SAPBULKEXTRACT
from pydantic import Field
from typing import ClassVar


class sapbulkextract(BaseStage):

    op_name: ClassVar[str] = "sapbulkextract"
    node_type: ClassVar[str] = "binding"
    image: ClassVar[str] = "/data-intg/flows/graphics/palette/sapbulkextract.svg"
    label: ClassVar[str] = "SAP Bulk Extract"
    runtime_column_propagation: bool = Field(False, alias="runtime_column_propagation")
    connection: SapbulkextractConn = SapbulkextractConn()
    buf_free_run_ronly: int | None = Field(50, alias="buf_free_run_ronly")
    buf_mode_ronly: SAPBULKEXTRACT.BufModeRonly | None = Field(
        SAPBULKEXTRACT.BufModeRonly.default, alias="buf_mode_ronly"
    )
    buffer_free_run_percent: int | None = Field(50, alias="buf_free_run")
    buffering_mode: SAPBULKEXTRACT.BufferingMode | None = Field(
        SAPBULKEXTRACT.BufferingMode.default, alias="buf_mode"
    )
    collecting: SAPBULKEXTRACT.Collecting | None = Field(SAPBULKEXTRACT.Collecting.auto, alias="coll_type")
    column_metadata_change_propagation: bool | None = Field(None, alias="auto_column_propagation")
    combinability_mode: SAPBULKEXTRACT.CombinabilityMode | None = Field(
        SAPBULKEXTRACT.CombinabilityMode.auto, alias="combinability"
    )
    condition_clauses: str | None = Field(None, alias="condition_clause")
    connection_count: int | None = Field(1, alias="connection_count")
    create_rfc_destination: bool | None = Field(True, alias="create_rfc_destination")
    create_rfc_destination_name: bool | None = Field(True, alias="create_rfc_destination_name")
    current_output_link_type: str = Field("PRIMARY", alias="currentOutputLinkType")
    db2_database_name: str | None = Field(None, alias="part_client_dbname")
    db2_instance_name: str | None = Field(None, alias="part_client_instance")
    db2_source_connection_required: str | None = Field("", alias="part_dbconnection")
    db2_table_name: str | None = Field(None, alias="part_table")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    disk_write_inc_ronly: int | None = Field(1048576, alias="disk_write_inc_ronly")
    disk_write_increment_bytes: int | None = Field(1048576, alias="disk_write_inc")
    ds_java_heap_size: int | None = Field(256, alias="_java._heap_size")
    enable_flow_acp_control: bool = Field(True, alias="enableFlowAcpControl")
    enable_schemaless_design: bool = Field(False, alias="enableSchemalessDesign")
    execution_mode: SAPBULKEXTRACT.ExecutionMode | None = Field(
        SAPBULKEXTRACT.ExecutionMode.default_par, alias="execmode"
    )
    extract_data_in_foreground: bool | None = Field(False, alias="extract_in_foreground")
    flow_dirty: str | None = Field("false", alias="flow_dirty")
    gateway_host: str | None = Field(None, alias="gateway_host")
    generate_sql_query: bool | None = Field(False, alias="generate_sql_query")
    generate_unicode_type_columns: bool | None = Field(False, alias="generate_unicode_columns")
    hide: bool | None = Field(False, alias="hide")
    input_count: int | None = Field(0, alias="input_count")
    input_link_description: list | None = Field("", alias="inputLinkDescription")
    inputcol_properties: list | None = Field([], alias="inputcolProperties")
    job: str | None = Field(None, alias="job")
    job_timeout_in_seconds: int | None = Field(600, alias="timeout")
    key_cols_coll: list | None = Field([], alias="keyColsColl")
    key_cols_coll_ordered: list | None = Field([], alias="keyColsCollOrdered")
    key_cols_coll_robin: list | None = Field([], alias="keyColsCollRobin")
    key_cols_none: list | None = Field([], alias="keyColsNone")
    key_cols_part: list | None = Field([], alias="keyColsPart")
    max_mem_buf_size_ronly: int | None = Field(3145728, alias="max_mem_buf_size_ronly")
    maximum_memory_buffer_size_bytes: int | None = Field(3145728, alias="max_mem_buf_size")
    metadata_table: str | None = Field(None, alias="metadata_table")
    node_count: int | None = Field(1, alias="node_count")
    node_number: int | None = Field(0, alias="node_number")
    output_acp_should_hide: bool = Field(True, alias="outputAcpShouldHide")
    output_count: int | None = Field(0, alias="output_count")
    output_link_description: list | None = Field("", alias="outputLinkDescription")
    outputcol_properties: list | None = Field([], alias="outputcolProperties")
    package_name: str | None = Field(None, alias="package_name")
    packet_size: int | None = Field(50000, alias="packet_size")
    part_stable_coll: bool | None = Field(False, alias="part_stable_coll")
    part_stable_ordered: bool | None = Field(False, alias="part_stable_ordered")
    part_stable_roundrobin_coll: bool | None = Field(False, alias="part_stable_roundrobin_coll")
    part_unique_coll: bool | None = Field(False, alias="part_unique_coll")
    part_unique_ordered: bool | None = Field(False, alias="part_unique_ordered")
    part_unique_roundrobin_coll: bool | None = Field(False, alias="part_unique_roundrobin_coll")
    partition_type: SAPBULKEXTRACT.PartitionType | None = Field(SAPBULKEXTRACT.PartitionType.auto, alias="part_type")
    perform_sort: bool | None = Field(False, alias="perform_sort")
    perform_sort_coll: bool | None = Field(False, alias="perform_sort_coll")
    perform_sort_modulus: bool | None = Field(False, alias="perform_sort_modulus")
    preserve_partitioning: SAPBULKEXTRACT.PreservePartitioning | None = Field(
        SAPBULKEXTRACT.PreservePartitioning.default_propagate, alias="preserve"
    )
    program_id: str | None = Field(None, alias="program_id")
    queue_upper_bound_size_bytes: int | None = Field(0, alias="queue_upper_size")
    queue_upper_size_ronly: int | None = Field(0, alias="queue_upper_size_ronly")
    rfc_destination_name: str | None = Field(None, alias="rfc_destination")
    row_limit: str | None = Field(None, alias="row_limit")
    runtime_column_propagation: bool | None = Field(None, alias="runtime_column_propagation")
    show_coll_type: int | None = Field(0, alias="showCollType")
    show_part_type: int | None = Field(1, alias="showPartType")
    show_sort_options: int | None = Field(0, alias="showSortOptions")
    sort_instructions: str | None = Field("", alias="sortInstructions")
    sort_instructions_text: str | None = Field("", alias="sortInstructionsText")
    sorting_key: SAPBULKEXTRACT.KeyColSelect | None = Field(
        SAPBULKEXTRACT.KeyColSelect.default, alias="keyColSelect"
    )
    sql_query: str | None = Field(None, alias="sql_query")
    stable: bool | None = Field(None, alias="part_stable")
    stage_description: list | None = Field("", alias="stageDescription")
    system_number_gateway_service: str | None = Field(None, alias="gateway_service")
    table_name: str | None = Field(None, alias="table_name")
    unique: bool | None = Field(None, alias="part_unique")
    use_secured_port: bool | None = Field(False, alias="use_secured_port")

    def validate_parameters(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        (
            include.add("preserve_partitioning")
            if (self.output_count and self.output_count > 0)
            else exclude.add("preserve_partitioning")
        )
        return include, exclude

    def validate_source(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        include.add("condition_clauses") if (self.generate_sql_query) else exclude.add("condition_clauses")
        (
            include.add("program_id")
            if (((not self.create_rfc_destination_name)) or (not self.create_rfc_destination))
            else exclude.add("program_id")
        )
        include.add("sql_query") if (not self.generate_sql_query) else exclude.add("sql_query")
        (
            include.add("create_rfc_destination_name")
            if (self.create_rfc_destination)
            else exclude.add("create_rfc_destination_name")
        )
        include.add("table_name") if (self.generate_sql_query) else exclude.add("table_name")
        (
            include.add("rfc_destination_name")
            if (((not self.create_rfc_destination_name)) or (not self.create_rfc_destination))
            else exclude.add("rfc_destination_name")
        )
        (
            include.add("condition_clauses")
            if (
                ((self.generate_sql_query))
                or (self.generate_sql_query and "#" in str(self.generate_sql_query))
            )
            else exclude.add("condition_clauses")
        )
        (
            include.add("program_id")
            if (
                ((not self.create_rfc_destination_name))
                or (not self.create_rfc_destination)
                or (self.create_rfc_destination_name and "#" in str(self.create_rfc_destination_name))
                or (self.create_rfc_destination and "#" in str(self.create_rfc_destination))
            )
            else exclude.add("program_id")
        )
        (
            include.add("sql_query")
            if (
                ((not self.generate_sql_query))
                or (self.generate_sql_query and "#" in str(self.generate_sql_query))
            )
            else exclude.add("sql_query")
        )
        (
            include.add("create_rfc_destination_name")
            if (
                ((self.create_rfc_destination))
                or (self.create_rfc_destination and "#" in str(self.create_rfc_destination))
            )
            else exclude.add("create_rfc_destination_name")
        )
        (
            include.add("table_name")
            if (
                ((self.generate_sql_query))
                or (self.generate_sql_query and "#" in str(self.generate_sql_query))
            )
            else exclude.add("table_name")
        )
        (
            include.add("rfc_destination_name")
            if (
                ((not self.create_rfc_destination_name))
                or (not self.create_rfc_destination)
                or (self.create_rfc_destination_name and "#" in str(self.create_rfc_destination_name))
                or (self.create_rfc_destination and "#" in str(self.create_rfc_destination))
            )
            else exclude.add("rfc_destination_name")
        )

        (
            include.add("maximum_memory_buffer_size_bytes")
            if (self.buffering_mode != "nobuffer")
            else exclude.add("maximum_memory_buffer_size_bytes")
        )
        (
            include.add("buffer_free_run_percent")
            if (self.buffering_mode != "nobuffer")
            else exclude.add("buffer_free_run_percent")
        )
        (
            include.add("queue_upper_bound_size_bytes")
            if (self.buffering_mode != "nobuffer")
            else exclude.add("queue_upper_bound_size_bytes")
        )
        (
            include.add("disk_write_increment_bytes")
            if (self.buffering_mode != "nobuffer")
            else exclude.add("disk_write_increment_bytes")
        )
        (
            include.add("runtime_column_propagation")
            if (not self.enable_schemaless_design)
            else exclude.add("runtime_column_propagation")
        )
        (
            include.add("column_metadata_change_propagation")
            if (not self.output_acp_should_hide)
            else exclude.add("column_metadata_change_propagation")
        )
        (
            include.add("max_mem_buf_size_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("max_mem_buf_size_ronly")
        )
        (
            include.add("buf_free_run_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("buf_free_run_ronly")
        )
        (
            include.add("queue_upper_size_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("queue_upper_size_ronly")
        )
        (
            include.add("disk_write_inc_ronly")
            if (self.buf_mode_ronly != "nobuffer")
            else exclude.add("disk_write_inc_ronly")
        )
        (
            include.add("stable")
            if (

                    ((self.show_part_type))
                    and (self.partition_type != "auto")
                    and (self.partition_type != "db2connector")
                    and (self.show_sort_options)

            )
            else exclude.add("stable")
        )
        (
            include.add("unique")
            if (

                    ((self.show_part_type))
                    and (self.partition_type != "auto")
                    and (self.partition_type != "db2connector")
                    and (self.show_sort_options)

            )
            else exclude.add("unique")
        )
        (
            include.add("key_cols_part")
            if (
                (
                    (
                        ((self.show_part_type))
                        and ((not self.show_coll_type))
                        and ((self.partition_type != "auto"))
                        and ((self.partition_type != "db2connector"))
                        and ((self.partition_type != "modulus"))
                    )
                )
                or (

                        ((self.show_part_type))
                        and (not self.show_coll_type)
                        and (self.partition_type == "modulus")
                        and (self.perform_sort_modulus)

                )
            )
            else exclude.add("key_cols_part")
        )
        (
            include.add("db2_source_connection_required")
            if (((self.partition_type == "db2part")) and (self.show_part_type))
            else exclude.add("db2_source_connection_required")
        )
        (
            include.add("db2_database_name")
            if (((self.partition_type == "db2part")) and (self.show_part_type))
            else exclude.add("db2_database_name")
        )
        (
            include.add("db2_instance_name")
            if (((self.partition_type == "db2part")) and (self.show_part_type))
            else exclude.add("db2_instance_name")
        )
        (
            include.add("db2_table_name")
            if (((self.partition_type == "db2part")) and (self.show_part_type))
            else exclude.add("db2_table_name")
        )
        (
            include.add("perform_sort")
            if (

                    ((self.show_part_type))
                    and (((self.partition_type == "hash")) or (self.partition_type == "range"))

            )
            else exclude.add("perform_sort")
        )
        (
            include.add("perform_sort_modulus")
            if (((self.show_part_type)) and (self.partition_type == "modulus"))
            else exclude.add("perform_sort_modulus")
        )
        (
            include.add("sorting_key")
            if (

                    ((self.show_part_type))
                    and (self.partition_type == "modulus")
                    and (not self.perform_sort_modulus)

            )
            else exclude.add("sorting_key")
        )
        (
            include.add("sort_instructions")
            if (

                    ((self.show_part_type))
                    and (
                        ((self.partition_type == "db2part"))
                        or (self.partition_type == "entire")
                        or (self.partition_type == "random")
                        or (self.partition_type == "roundrobin")
                        or (self.partition_type == "same")
                    )

            )
            else exclude.add("sort_instructions")
        )
        (
            include.add("sort_instructions_text")
            if (

                    ((self.show_part_type))
                    and (
                        ((self.partition_type == "db2part"))
                        or (self.partition_type == "entire")
                        or (self.partition_type == "random")
                        or (self.partition_type == "roundrobin")
                        or (self.partition_type == "same")
                    )

            )
            else exclude.add("sort_instructions_text")
        )
        include.add("collecting") if (self.show_coll_type) else exclude.add("collecting")
        include.add("partition_type") if (self.show_part_type) else exclude.add("partition_type")
        (
            include.add("perform_sort_coll")
            if (
                (
                    (
                        ((self.show_coll_type))
                        and (
                            ((self.collecting == "ordered"))
                            or ((self.collecting == "roundrobin_coll"))
                            or ((self.collecting == "sortmerge"))
                        )
                    )
                )
                or (((not self.show_part_type)) and (not self.show_coll_type))
            )
            else exclude.add("perform_sort_coll")
        )
        (
            include.add("key_cols_coll")
            if (

                    ((self.show_coll_type))
                    and (not self.show_part_type)
                    and (self.collecting != "auto")
                    and (((self.collecting == "sortmerge")) or (self.perform_sort_coll))

            )
            else exclude.add("key_cols_coll")
        )
        (
            include.add("key_cols_none")
            if (

                    ((not self.show_part_type))
                    and (not self.show_coll_type)
                    and (self.perform_sort_coll)

            )
            else exclude.add("key_cols_none")
        )
        (
            include.add("part_stable_coll")
            if (

                    (((self.perform_sort_coll)))
                    and (
                        (
                            (
                                ((not self.show_part_type))
                                and ((self.show_coll_type))
                                and ((self.collecting != "auto"))
                                and ((self.show_sort_options))
                            )
                        )
                        or (

                                ((not self.show_part_type))
                                and (not self.show_coll_type)
                                and (self.show_sort_options)

                        )
                    )

            )
            else exclude.add("part_stable_coll")
        )
        (
            include.add("part_unique_coll")
            if (

                    (((self.perform_sort_coll)))
                    and (
                        (
                            (
                                ((not self.show_part_type))
                                and ((self.show_coll_type))
                                and ((self.collecting != "auto"))
                                and ((self.show_sort_options))
                            )
                        )
                        or (

                                ((not self.show_part_type))
                                and (not self.show_coll_type)
                                and (self.show_sort_options)

                        )
                    )

            )
            else exclude.add("part_unique_coll")
        )
        include.add("defer_credentials") if (()) else exclude.add("defer_credentials")
        (
            include.add("create_rfc_destination_name")
            if (self.create_rfc_destination == "true" or self.create_rfc_destination)
            else exclude.add("create_rfc_destination_name")
        )
        (
            include.add("rfc_destination_name")
            if (self.create_rfc_destination_name == "false" or not self.create_rfc_destination_name)
            or (self.create_rfc_destination == "false" or not self.create_rfc_destination)
            else exclude.add("rfc_destination_name")
        )
        (
            include.add("table_name")
            if (self.generate_sql_query == "true" or self.generate_sql_query)
            else exclude.add("table_name")
        )
        (
            include.add("program_id")
            if (self.create_rfc_destination_name == "false" or not self.create_rfc_destination_name)
            or (self.create_rfc_destination == "false" or not self.create_rfc_destination)
            else exclude.add("program_id")
        )
        (
            include.add("condition_clauses")
            if (self.generate_sql_query == "true" or self.generate_sql_query)
            else exclude.add("condition_clauses")
        )
        (
            include.add("sql_query")
            if (self.generate_sql_query == "false" or not self.generate_sql_query)
            else exclude.add("sql_query")
        )
        return include, exclude

    def get_source_props(self) -> dict:
        include, exclude = self.validate_source()
        props = {
            "buf_free_run_ronly",
            "buf_mode_ronly",
            "buffer_free_run_percent",
            "buffering_mode",
            "collecting",
            "column_metadata_change_propagation",
            "combinability_mode",
            "condition_clauses",
            "connection_count",
            "create_rfc_destination",
            "create_rfc_destination_name",
            "current_output_link_type",
            "db2_database_name",
            "db2_instance_name",
            "db2_source_connection_required",
            "db2_table_name",
            "disk_write_inc_ronly",
            "disk_write_increment_bytes",
            "ds_java_heap_size",
            "enable_flow_acp_control",
            "enable_schemaless_design",
            "execution_mode",
            "extract_data_in_foreground",
            "flow_dirty",
            "gateway_host",
            "generate_sql_query",
            "generate_unicode_type_columns",
            "hide",
            "input_count",
            "input_link_description",
            "inputcol_properties",
            "job",
            "job_timeout_in_seconds",
            "key_cols_coll",
            "key_cols_coll_ordered",
            "key_cols_coll_robin",
            "key_cols_none",
            "key_cols_part",
            "max_mem_buf_size_ronly",
            "maximum_memory_buffer_size_bytes",
            "metadata_table",
            "node_count",
            "node_number",
            "output_acp_should_hide",
            "output_count",
            "output_link_description",
            "outputcol_properties",
            "package_name",
            "packet_size",
            "part_stable_coll",
            "part_stable_ordered",
            "part_stable_roundrobin_coll",
            "part_unique_coll",
            "part_unique_ordered",
            "part_unique_roundrobin_coll",
            "partition_type",
            "perform_sort",
            "perform_sort_coll",
            "perform_sort_modulus",
            "preserve_partitioning",
            "program_id",
            "queue_upper_bound_size_bytes",
            "queue_upper_size_ronly",
            "rfc_destination_name",
            "row_limit",
            "runtime_column_propagation",
            "show_coll_type",
            "show_part_type",
            "show_sort_options",
            "sort_instructions",
            "sort_instructions_text",
            "sorting_key",
            "sql_query",
            "stable",
            "stage_description",
            "system_number_gateway_service",
            "table_name",
            "unique",
            "use_secured_port",
        }
        required = {
            "application_server",
            "client_number",
            "connection_type",
            "current_output_link_type",
            "enable_flow_acp_control",
            "enable_schemaless_design",
            "group",
            "language",
            "message_server",
            "output_acp_should_hide",
            "sap_application_system_number",
            "snc_name",
            "snc_partner_name",
            "system_id",
            "system_number",
            "username",
            "x_509_certificate",
        }
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_parameters_props(self) -> dict:
        include, exclude = self.validate_parameters()
        props = {"execution_mode", "input_count", "output_count", "preserve_partitioning"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_app_data_props(self) -> dict:
        return {
            "datastage": {
                "active": 0,
                "SupportsRef": True,
                "maxRejectOutputs": 0,
                "minRejectOutputs": 0,
                "maxReferenceInputs": 0,
                "minReferenceInputs": 0,
            }
        }

    def get_input_ports_props(self) -> dict:
        include, exclude = self.validate_target()
        props = {"runtime_column_propagation"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_input_cardinality(self) -> dict:
        return {"min": 0, "max": 0}

    def get_output_ports_props(self) -> dict:
        include, exclude = self.validate_source()
        props = set()
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_output_cardinality(self) -> dict:
        return {"min": 0, "max": 1}

    def get_allowed_as_source_props(self) -> bool:
        return True

    def get_target_props(self) -> dict:
        include, exclude = self.validate_target()
        props = {}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import SALESFORCEAPI_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthenticationType(Enum):
    oauth_jwt = "oauth_jwt"
    oauth_username_and_password = "oauth_username_and_password"
    username_and_password = "username_and_password"


class SalesforceapiConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "3a00dbd2-2540-4976-afc2-5fc59f68ed35"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    authentication_type: SALESFORCEAPI_CONNECTION.AuthenticationType | None = Field(
        SALESFORCEAPI_CONNECTION.AuthenticationType.username_and_password, alias="authentication_type"
    )
    cas_lite_service_authorization_header: str | None = Field(None, alias="cas_lite_auth_header")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    consumer_key: str | None = Field(None, alias="consumer_key")
    consumer_secret_key: str | None = Field(None, alias="consumer_secret_key")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    password: str = Field(None, alias="password")
    proxy_server: bool | None = Field(False, alias="proxy_server")
    proxy_server_hostname_or_ip_address: str = Field(None, alias="proxy_server_host")
    proxy_server_password: str | None = Field(None, alias="proxy_server_password")
    proxy_server_port: str = Field(None, alias="proxy_server_port")
    proxy_server_username: str | None = Field(None, alias="proxy_server_username")
    schema_name: str | None = Field(None, alias="schema_name")
    server_certificate_key: str | None = Field(None, alias="server_certificate")
    token_expiry_time: int | None = Field(None, alias="token_expiry_time")
    url: str = Field(None, alias="url")
    use_cas_lite_service: bool | None = Field(True, alias="use_cas_lite")
    username: str = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        (
            include.add("consumer_key")
            if (
                (
                    (
                        self.authentication_type
                        and (
                            (
                                hasattr(self.authentication_type, "value")
                                and self.authentication_type.value == "oauth_jwt"
                            )
                            or (self.authentication_type == "oauth_jwt")
                        )
                    )
                )
                or (
                    self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value == "oauth_username_and_password"
                        )
                        or (self.authentication_type == "oauth_username_and_password")
                    )
                )
            )
            else exclude.add("consumer_key")
        )
        (
            include.add("server_certificate_key")
            if (
                self.authentication_type
                and (
                    (hasattr(self.authentication_type, "value") and self.authentication_type.value == "oauth_jwt")
                    or (self.authentication_type == "oauth_jwt")
                )
            )
            else exclude.add("server_certificate_key")
        )
        (
            include.add("password")
            if (
                ((not self.defer_credentials))
                and (
                    (
                        (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value == "oauth_username_and_password"
                                )
                                or (self.authentication_type == "oauth_username_and_password")
                            )
                        )
                    )
                    or (
                        self.authentication_type
                        and (
                            (
                                hasattr(self.authentication_type, "value")
                                and self.authentication_type.value == "username_and_password"
                            )
                            or (self.authentication_type == "username_and_password")
                        )
                    )
                )
            )
            else exclude.add("password")
        )
        include.add("proxy_server_password") if (self.proxy_server) else exclude.add("proxy_server_password")
        (
            include.add("consumer_secret_key")
            if (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value == "oauth_username_and_password"
                    )
                    or (self.authentication_type == "oauth_username_and_password")
                )
            )
            else exclude.add("consumer_secret_key")
        )
        include.add("proxy_server_username") if (self.proxy_server) else exclude.add("proxy_server_username")
        (
            include.add("use_cas_lite_service")
            if (self.cas_lite_service_authorization_header)
            else exclude.add("use_cas_lite_service")
        )
        (
            include.add("cas_lite_service_authorization_header")
            if (self.use_cas_lite_service)
            else exclude.add("cas_lite_service_authorization_header")
        )
        include.add("proxy_server_port") if (self.proxy_server) else exclude.add("proxy_server_port")
        (
            include.add("proxy_server_hostname_or_ip_address")
            if (self.proxy_server)
            else exclude.add("proxy_server_hostname_or_ip_address")
        )
        include.add("username") if (not self.defer_credentials) else exclude.add("username")
        (
            include.add("token_expiry_time")
            if (
                self.authentication_type
                and (
                    (hasattr(self.authentication_type, "value") and self.authentication_type.value == "oauth_jwt")
                    or (self.authentication_type == "oauth_jwt")
                )
            )
            else exclude.add("token_expiry_time")
        )
        (
            include.add("password")
            if (
                (
                    (
                        self.authentication_type
                        and (
                            (
                                hasattr(self.authentication_type, "value")
                                and self.authentication_type.value == "oauth_username_and_password"
                            )
                            or (self.authentication_type == "oauth_username_and_password")
                        )
                    )
                )
                or (
                    self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value == "username_and_password"
                        )
                        or (self.authentication_type == "username_and_password")
                    )
                )
            )
            else exclude.add("password")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )

        (
            include.add("server_certificate_key")
            if (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "oauth_jwt" in str(self.authentication_type.value)
                    )
                    or ("oauth_jwt" in str(self.authentication_type))
                )
            )
            else exclude.add("server_certificate_key")
        )
        (
            include.add("use_cas_lite_service")
            if (self.cas_lite_service_authorization_header)
            else exclude.add("use_cas_lite_service")
        )
        (
            include.add("cas_lite_service_authorization_header")
            if (self.use_cas_lite_service == "true" or self.use_cas_lite_service)
            else exclude.add("cas_lite_service_authorization_header")
        )
        (
            include.add("token_expiry_time")
            if (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "oauth_jwt" in str(self.authentication_type.value)
                    )
                    or ("oauth_jwt" in str(self.authentication_type))
                )
            )
            else exclude.add("token_expiry_time")
        )
        (
            include.add("proxy_server_port")
            if (self.proxy_server and "true" in str(self.proxy_server))
            else exclude.add("proxy_server_port")
        )
        (
            include.add("proxy_server_password")
            if (self.proxy_server and "true" in str(self.proxy_server))
            else exclude.add("proxy_server_password")
        )
        (
            include.add("password")
            if (not self.defer_credentials)
            and (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "oauth_username_and_password" in str(self.authentication_type.value)
                    )
                    or ("oauth_username_and_password" in str(self.authentication_type))
                )
                or self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "username_and_password" in str(self.authentication_type.value)
                    )
                    or ("username_and_password" in str(self.authentication_type))
                )
            )
            else exclude.add("password")
        )
        include.add("username") if not self.defer_credentials else exclude.add("username")
        (
            include.add("proxy_server_username")
            if (self.proxy_server and "true" in str(self.proxy_server))
            else exclude.add("proxy_server_username")
        )
        (
            include.add("consumer_key")
            if (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "oauth_jwt" in str(self.authentication_type.value)
                    )
                    or ("oauth_jwt" in str(self.authentication_type))
                )
                or self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "oauth_username_and_password" in str(self.authentication_type.value)
                    )
                    or ("oauth_username_and_password" in str(self.authentication_type))
                )
            )
            else exclude.add("consumer_key")
        )
        (
            include.add("proxy_server_hostname_or_ip_address")
            if (self.proxy_server and "true" in str(self.proxy_server))
            else exclude.add("proxy_server_hostname_or_ip_address")
        )
        (
            include.add("consumer_secret_key")
            if (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "oauth_username_and_password" in str(self.authentication_type.value)
                    )
                    or ("oauth_username_and_password" in str(self.authentication_type))
                )
            )
            else exclude.add("consumer_secret_key")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing Engine Models."""

from copy import deepcopy
from functools import cached_property
from ibm_watsonx_data_integration.common.models import BaseModel
from ibm_watsonx_data_integration.services.streamsets.api.sdc_api import DataCollectorAPIClient
from pydantic import ConfigDict, Field, PrivateAttr
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.cpd_models import Project
    from ibm_watsonx_data_integration.platform import Platform


class EngineMetrics(BaseModel):
    """StreamSets Engine Metrics."""

    cpu_load: float = Field(repr=True)
    memory_load: float = Field(repr=True)
    job_count: int = Field(repr=True)

    model_config = ConfigDict(frozen=True)


class EngineHealthMetadata(BaseModel):
    """StreamSets Engine Health Metadata."""

    status: str = Field(frozen=True, repr=True)
    last_heartbeat_time: int | None = Field(frozen=True, repr=False, default=None)
    last_status_updated_time: int | None = Field(repr=False, default=None)
    last_startup_time: int | None = Field(repr=False, default=None)
    last_registration_time: int | None = Field(repr=False, default=None)
    metrics: EngineMetrics | None = Field(repr=True, default=None)

    model_config = ConfigDict(frozen=True)


class EngineMetadata(BaseModel):
    """The model for CPD StreamSets Engine Metadata."""

    name: str = Field(repr=False)
    description: str = Field(repr=False)
    owner: str = Field(repr=False, default=None)
    owner_id: str = Field(repr=False)
    created: int = Field(repr=False)
    created_at: str = Field(repr=False)

    tags: list | None = Field(repr=False, default_factory=list)

    project_id: str = Field(repr=False)
    asset_id: str = Field(repr=True)
    asset_attributes: list | None = Field(repr=False)
    asset_state: str | None = Field(repr=False)
    asset_type: str | None = Field(repr=False)
    origin_country: str | None = Field(repr=False, default=None)
    catalog_id: str = Field(repr=False)
    rating: int | None = Field(repr=False)
    size: int | None = Field(repr=False)
    space_id: str = Field(repr=False, default=None)
    rov: dict | None = Field(repr=False)
    usage: dict | None = Field(repr=False)
    version: int | None = Field(repr=False)

    create_time: str | None = Field(repr=False, default=None)
    sandbox_id: str = Field(repr=False)
    creator_id: str = Field(repr=False)

    model_config = ConfigDict(frozen=True)
    _expose: bool = PrivateAttr(default=False)


class LibraryDefinitions(BaseModel):
    """An engine's library definition."""

    _expose: bool = PrivateAttr(default=False)

    schema_version: str = Field(alias="schemaVersion", repr=False)
    pipeline: list[dict] = Field(alias="pipeline", repr=False)
    pipeline_fragment: list[dict] = Field(alias="pipelineFragment", repr=False)
    pipeline_rules: list[dict] = Field(alias="pipelineRules", repr=False)
    stages: list[dict] = Field(alias="stages", repr=False)
    services: list[dict] = Field(alias="services", repr=False)
    rules_el_metadata: dict = Field(alias="rulesElMetadata", repr=False)
    el_catalog: Any = Field(alias="elCatalog", repr=False)
    runtime_configs: list[Any] = Field(alias="runtimeConfigs", repr=False)
    stage_icons: Any = Field(alias="stageIcons", repr=False)
    legacy_stage_libs: list[dict] = Field(alias="legacyStageLibs", repr=False)
    event_definitions: dict = Field(alias="eventDefinitions", repr=False)
    version: int | None = Field(alias="version", repr=False)
    executor_version: str | None = Field(alias="executorVersion", repr=False)
    category: Any | None = Field(alias="category", repr=False)
    category_label: str | None = Field(alias="categoryLabel", repr=False)
    stage_definition_map: dict | None = Field(alias="stageDefinitionMap", repr=False)
    stage_definition_minimal_list: list[dict] | None = Field(alias="stageDefinitionMinimalList", repr=False)

    def __init__(self, **library_definitions: any) -> None:
        """The __init__ for this class."""
        processed_definitions = self.process_library_definitions(library_definitions=library_definitions)
        super().__init__(**processed_definitions)

    @staticmethod
    def process_library_definitions(library_definitions: dict) -> dict:
        """Copied from https://shorturl.at/sfrgJ.

        This entire function essentially expands on stageDefinitionMap and stageDefinitionMinimalList
        to make the entire library definitions.

        This is only required for library definitions gotten from CPD and not from an engine as it is compressed \
        to contain all possible stage libraries.
        """
        processed_definitions = deepcopy(library_definitions)
        processed_definitions["stageDefinitionMap"] = dict()

        for key, stage_definition in library_definitions.get("stageDefinitionMap").items():
            processed_definitions["stageDefinitionMap"][key] = deepcopy(stage_definition)
            processed_definitions["stageDefinitionMap"][key]["services"] = list()

        processed_definitions["stages"] = list()
        for stage_definition_minimal in library_definitions.get("stageDefinitionMinimalList"):
            key = f"{stage_definition_minimal['name']}::{stage_definition_minimal['version']}"
            if key in library_definitions.get("stageDefinitionMap", {}):
                stage_definition = deepcopy(library_definitions.get("stageDefinitionMap")[key])
                stage_definition["library"] = stage_definition_minimal["library"]
                stage_definition["libraryLabel"] = stage_definition_minimal["libraryLabel"]
                processed_definitions["stages"].append(stage_definition)

        if not processed_definitions.get("injectedServices"):
            services = processed_definitions["services"]
            processed_definitions["injectedServices"] = True

            services_name_map = {}
            for service in services:
                services_name_map[service["provides"]] = service

            for key, stage_definition in library_definitions.get("stageDefinitionMap").items():
                if stage_definition.get("services"):
                    processed_definitions["stageDefinitionMap"][key]["services"] = []
                    for service in stage_definition.get("services"):
                        service_definition = deepcopy(services_name_map[service["service"]])
                        service_definition["configuration"] = service["configuration"]
                        processed_definitions["stageDefinitionMap"][key]["services"].append(service_definition)

        return processed_definitions

    @classmethod
    def for_engine_version(cls, platform: "Platform", engine_version: str) -> "LibraryDefinitions":
        """Returns the library definitions for a particular engine version, includes all possible stage libraries.

        Args:
            platform: Instance of platform to get the definitions from.
            engine_version: The engine version for which library definitions need to be fetched.

        Returns:
            An instance of the LibraryDefinitions class.
        """
        return cls(
            **platform._environment_api.get_library_definitions_for_engine_version(engine_version=engine_version).json()
        )


class Engine(BaseModel):
    """The Model for Engine."""

    EXPOSED_DATA_PATH = {
        "entity.streamsets_engine": {},
    }

    metadata: EngineMetadata = Field(repr=True)

    registration_status: str | None = Field(repr=False)
    registration_time: int = Field(repr=False)
    reported_engine_version: str = Field(repr=False, default=None)
    reported_build_time: int = Field(repr=False, default=None)
    reported_build_sha: str = Field(repr=False, default=None)
    reported_java_vendor: str | None = Field(repr=False, default=None)
    reported_java_version: str = Field(repr=False, default=None)
    reported_os_name: str = Field(repr=False, default=None)
    reported_os_arch: str = Field(repr=False, default=None)
    reported_os_version: str = Field(repr=False, default=None)
    last_startup_time: int = Field(repr=True, default=None)
    engine_type: str = Field(repr=True)
    url: str = Field(repr=True, default=None)
    streamsets_environment_asset_id: str = Field(repr=False)
    health: EngineHealthMetadata | None = Field(repr=False, default=None)

    model_config = ConfigDict(frozen=True)

    def __init__(self, platform: "Platform" = None, project: "Project" = None, **engine_json: dict) -> None:
        """The __init__ of the Engine class.

        Args:
            platform: The Platform object.
            project: The Project object.
            engine_json: The JSON for the Engine.
        """
        super().__init__(**engine_json)
        self._platform = platform
        self._project = project

    def __str__(self) -> str:
        """Custom __str__ to include the product property.

        Returns:
            A string representation of class.
        """
        return f"Engine(name={self.metadata.name}, engine_id={self.metadata.asset_id}, \
            project_id={self.metadata.project_id})"

    @property
    def engine_id(self) -> str:
        """Returns engine asset_id."""
        return self.metadata.asset_id

    @cached_property
    def api_client(self) -> DataCollectorAPIClient:
        """The API Client connected directly to the engine."""
        return DataCollectorAPIClient(auth=self._platform._engine_api._auth, engine_url=self.url)

    @cached_property
    def library_definitions(self) -> LibraryDefinitions:
        """Library Definitions of the Engine."""
        return LibraryDefinitions(**self.api_client.get_library_definitions().json())



from ibm_watsonx_data_integration.services.datastage.models.enums import ORACLE_CONNECTION
from ibm_watsonx_data_integration.services.datastage.models.flow.datastage_flow import DataStageFlow
from ibm_watsonx_data_integration.services.datastage.tests.integration.util import AbstractFlowRunTest
from os import environ
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestOracleConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "oracle_peek"

    def get_fc(self, project: Project):
        fc = DataStageFlow(project=project, name=self.get_flow_name(), description="", flow_type="datastage")

        # Stages
        _fastpathtest_1_ = fc.add_stage("Oracle", "FASTPATHTEST_1")
        _fastpathtest_1_.configuration.schema_name = "TM_DS"
        _fastpathtest_1_.configuration.table_name = "FASTPATHTEST"

        _fastpathtest_1_.configuration.connection.connection_mode = "service_name"
        _fastpathtest_1_.configuration.connection.hostname_or_ip_address = environ["ORACLE_HOST"]
        _fastpathtest_1_.configuration.connection.metadata_discovery = ORACLE_CONNECTION.MetadataDiscovery.no_remarks
        _fastpathtest_1_.configuration.connection.password = environ["ORACLE_PASSWORD"]
        _fastpathtest_1_.configuration.connection.port = "1521"
        _fastpathtest_1_.configuration.connection.proxy = False
        _fastpathtest_1_.configuration.connection.service_name = "orclpdb"
        _fastpathtest_1_.configuration.connection.port_is_ssl_enabled = False
        _fastpathtest_1_.configuration.connection.username = environ["ORACLE_USERNAME"]

        _peek_2_ = fc.add_stage("Peek", "Peek_2")

        # Graph
        fastpathtest_1_schema = _fastpathtest_1_.connect_output_to(_peek_2_).set_name("Link_2").create_schema()

        # Schemas
        fastpathtest_1_schema.add_field("CHAR", "NAME").nullable().length(6)
        fastpathtest_1_schema.add_field("BIGINT", "SAVINGS").nullable()
        fastpathtest_1_schema.add_field("BIGINT", "DEBT").nullable()

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



# -------------------------------------------------------------------
# This file was auto-generated. Do not edit directly.
# Generated by: StageGeneratorV2
# -------------------------------------------------------------------

from typing import TYPE_CHECKING

from ibm_watsonx_data_integration.services.datastage.models.flow.dag import StageNode
from ibm_watsonx_data_integration.services.datastage.models.stage_models.match_frequency import match_frequency


if TYPE_CHECKING:
    from ibm_watsonx_data_integration.services.datastage.models.flow.dag import DAG


# from ibm.datastage._framework.components.match_specification import MatchSpecification


class MatchFrequencyStage(StageNode):
    def __init__(self, dag: "DAG", configuration: match_frequency=None, *, label: str = None):
        super().__init__(dag, configuration=configuration, label=label)
        self.configuration: match_frequency = configuration or match_frequency()


    # def set_match_specification(self, match_specification: MatchSpecification):
    #     self.configuration.match_specification = match_specification
    #     self.configuration.specfile = f"{match_specification.name}.MAT"
    #     self.configuration.not_use_specfile = False
    #     self.configuration.workdir = (
    #         f"quality_stage/match_specification/{match_specification.name}"
    #     )
    #     self.configuration.maxfreq = match_specification.MAXFREQ
    #     self.configuration.match_asset_id = match_specification.asset_id
    #     self.configuration.match_type = match_specification.MATCHTYPE



CONN_MAPPINGS = {
    "amazons3": "amazons3",
    "AmazonS3PX": "amazons3",
    "apachekafka": "apache_kafka",
    "KafkaConnectorPX": "apache_kafka",
    "awsglue": "awsglue",
    "azureblobstorage": "azure_blob_storage",
    "azuredatalake": "azuredatalake",
    "AzureDatalakePX": "azuredatalake",
    "azurefilestorage": "azure_file_storage",
    "azuresql": "azuresql",
    "azuresynapse": "azuresynapse",
    "bigquery": "bigquery",
    "bigqueryPX": "bigquery",
    "bigsql": "bigsql",
    "bluemixcloudobjectstorage": "bluemixcloudobjectstorage",
    "box": "box",
    "cassandra": "cassandra",
    "CassandraConnectorPX": "cassandra",
    "cassandra-datastage": "cassandra_datastage",
    "cloudant": "cloudant",
    "cloudobjectstorage": "cloud_object_storage",
    "cloudobjectstoragePX": "cloud_object_storage",
    "cognos-analytics": "cognos_analytics",
    "cognos-query-service": "cognos-query-service",
    "cosmos": "azure_cosmos",
    "custom-noop": "custom-noop",
    "dashdb": "db2warehouse",
    "databricks": "azure_databricks",
    "datastax": "datastax",
    "db2": "db2",
    "DB2ConnectorPX": "db2",
    "db2-datastage": "db2fordatastage",
    "db2cloud": "db2cloud",
    "db2iseries": "db2iseries",
    "db2zos": "db2zos",
    "denodo": "denodo",
    "derby": "derby",
    "dremio": "dremio",
    "dropbox": "dropbox",
    "dv": "dv",
    "dvm": "dvm",
    "elastic-cloud": "elastic-cloud",
    "elasticsearch": "elasticsearch",
    "exasol": "exasol",
    "ftp": "ftp",
    "genericjdbc": "jdbc",
    "JDBCConnectorPX": "jdbc",
    "generics3": "generics3",
    "github": "github",
    "googlecloudstorage": "google_cloud_storage",
    "GoogleCloudStoragePX": "google_cloud_storage",
    "googlepubsub": "google_pub_sub",
    "GooglePubSubPX": "google_pub_sub",
    "greenplum": "greenplum",
    "hbase-datastage": "apache_hbase",
    "HBaseConnectorPX": "apache_hbase",
    "hdfs-apache": "hdfs_apache",
    "hive": "apache_hive",
    "HiveConnectorPX": "apache_hive",
    "http": "http",
    "ibm-datastage-cp4d": "ibm-datastage-cp4d",
    "ibm-infosphere-datastage": "ibm-infosphere-datastage",
    "impala": "impala",
    "informix": "informix",
    "lakehouse": "watsonx_data",
    "looker": "google_looker",
    "mariadb": "mariadb",
    "match360": "match360",
    "microstrategy": "microstrategy",
    "milvus": "milvus",
    "milvuswxd": "milvuswxd",
    "minio": "minio",
    "mongodb": "mongodb",
    "mongodb-ibmcloud": "mongodb_ibmcloud",
    "mysql": "mysql",
    "mysql-amazon": "mysql_amazon",
    "mysql-ibmcloud": "mysql_compose",
    "mysql-compose": "mysql_compose",
    "nearbysearch": "nearbysearch",
    "netezza": "netezza",
    "NetezzaConnectorPX": "netezza",
    "netezza-datastage": "netezza_optimized",
    "odata": "odata",
    "odbc-datastage": "odbc",
    "ODBCConnectorPX": "odbc",
    "openlineage": "openlineage",
    "oracle": "oracle",
    "OracleConnectorPX": "oracle",
    "oracle-amazon": "amazonrds_oracle",
    "oracle-datastage": "oracle_datastage",
    "postgresql": "postgresql",
    "postgresql-amazon": "amazon_postgresql",
    "postgresql-azure": "azure_postgresql",
    "postgresql-ibmcloud": "postgresql_ibmcloud",
    "powerbi-azure": "powerbi-azure",
    "presto": "presto",
    "redshift": "amazon_redshift",
    "RedshiftPX": "amazon_redshift",
    "restgeneric": "restgeneric",
    "riakkv": "riakkv",
    "salesforce": "salesforce",
    "salesforce-datastage": "salesforceapi",
    "SALESFORCEJCConnectorPX": "salesforceapi",
    "sapbapi": "sapbapi",
    "sapbulkextract": "sapbulkextract",
    "sapdeltaextract": "sapdeltaextract",
    "saphana": "saphana",
    "sapidoc": "sapidoc",
    "sapodata": "sapodata",
    "singlestore": "singlestore",
    "snowflake": "snowflake",
    "SnowflakeConnectorPX": "snowflake",
    "sqlserver": "sqlserver",
    "sybase": "sybase",
    "sybaseiq": "sapiq",
    "tableau": "tableau",
    "teradata": "teradata",
    "TeradataConnectorPX": "teradata",
    "teradata-datastage": "teradata_datastage",
    "tm1odata": "planning_analytics",
    "twitter": "twitter",
    "vertica": "vertica",
    "volumes": "storage_volume",
    "webspheremq-datastage": "ibm_mq",
    "WebSphereMQConnectorPX": "ibm_mq",
}
